from __future__ import annotations

import logging
import time
from typing import Optional, ClassVar, Dict, Sequence

import pandas as pd
import sqlalchemy
from databricks import sql

from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SqlEngineAttributes, SqlEngine
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer, SqlQueryPlanRenderer
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_clients.base_sql_client_implementation import BaseSqlClientImplementation
from metricflow.sql_clients.common_client import SqlDialect

logger = logging.getLogger(__name__)

HTTP_PATH_KEY = "httppath="

# This is a non-exhaustive list of pandas dtypes, but in theory it will cover the ones we need to support
# for data frames generated and run through type inference.
PANDAS_TO_SQL_DTYPES = {
    "string": "string",
    "object": "string",
    "float64": "double",
    "bool": "boolean",
    "boolean": "boolean",
    "int64": "int",
    "datetime64[ns]": "timestamp",
}


class DatabricksEngineAttributes(SqlEngineAttributes):
    """SQL engine attributes for Databricks."""

    sql_engine_type: ClassVar[SqlEngine] = SqlEngine.DATABRICKS

    # SQL Engine capabilities
    date_trunc_supported: ClassVar[bool] = True
    full_outer_joins_supported: ClassVar[bool] = True
    indexes_supported: ClassVar[bool] = False
    multi_threading_supported: ClassVar[bool] = True
    timestamp_type_supported: ClassVar[bool] = True
    timestamp_to_string_comparison_supported: ClassVar[bool] = True
    # So far the only clear way to cancel a query is through the Databricks UI.
    cancel_submitted_queries_supported: ClassVar[bool] = False

    # SQL Dialect replacement strings
    double_data_type_name: ClassVar[str] = "DOUBLE"
    timestamp_type_name: ClassVar[Optional[str]] = "TIMESTAMP"

    # MetricFlow attributes
    sql_query_plan_renderer: ClassVar[SqlQueryPlanRenderer] = DefaultSqlQueryPlanRenderer()


class DatabricksSqlClient(BaseSqlClientImplementation):
    """Client used to connect to Databricks engine."""

    def __init__(self, host: str, http_path: str, access_token: str) -> None:  # noqa: D
        self.host = host
        self.http_path = http_path
        self.access_token = access_token

    @staticmethod
    def from_connection_details(url: str, password: Optional[str]) -> DatabricksSqlClient:  # noqa: D
        """Parse MF_SQL_ENGINE_URL & MF_SQL_ENGINE_PASSWORD into useful connection params.

        Using just these 2 env variables ensures uniformity across engines.
        """
        try:
            split_url = url.split(";")
            parsed_url = sqlalchemy.engine.url.make_url(split_url[0])
            http_path = ""
            for piece in split_url[1:]:
                if HTTP_PATH_KEY in piece.lower():
                    __, http_path = piece.split("=")
                    break
            dialect = SqlDialect.DATABRICKS.value
            if not http_path:
                raise ValueError("HTTP path not found in URL.")
            if parsed_url.drivername != dialect:
                raise ValueError(f"Unexpected dialect in URL: {parsed_url.drivername}. Expected: {dialect}")
            if not parsed_url.host:
                raise ValueError("Host not found in URL.")
        except ValueError as e:
            # If any errors in parsing URL, show user what expected URL looks like.
            raise ValueError(
                f"Unexpected format for MF_SQL_ENGINE_URL. Expected: `{dialect}://<HOST>:443;HttpPath=<HTTP PATH>`."
            ) from e

        if not password:
            raise ValueError(f"Password not supplied for {url}")

        return DatabricksSqlClient(host=parsed_url.host, http_path=http_path, access_token=password)

    def get_connection(self) -> sql.client.Connection:
        """Get connection to Databricks cluster/warehouse."""
        return sql.connect(server_hostname=self.host, http_path=self.http_path, access_token=self.access_token)

    @property
    def sql_engine_attributes(self) -> SqlEngineAttributes:
        """Databricks engine attributes."""
        return DatabricksEngineAttributes()

    @staticmethod
    def params_or_none(bind_params: SqlBindParameters) -> Optional[Dict[str, str]]:
        """If there are no parameters, use None to prevent collision with `%` wildcard."""
        return None if bind_params == SqlBindParameters() else bind_params.param_dict

    def _engine_specific_query_implementation(self, stmt: str, bind_params: SqlBindParameters) -> pd.DataFrame:
        with self.get_connection() as connection:
            with connection.cursor() as cursor:
                cursor.execute(operation=stmt, parameters=self.params_or_none(bind_params))
                logger.info("Fetching query results as PyArrow Table.")
                pyarrow_df = cursor.fetchall_arrow()

        logger.info("Beginning conversion of PyArrow Table to pandas DataFrame.")
        pandas_df = pyarrow_df.to_pandas()
        logger.info("Completed conversion of PyArrow Table to pandas DataFrame.")
        # Remove tz from any datetime cols. Databricks tables add UTC by default.
        for col_name in pandas_df:
            if pd.api.types.is_datetime64_any_dtype(pandas_df[col_name]):
                pandas_df[col_name] = pandas_df[col_name].dt.tz_localize(None)
        return pandas_df

    def _engine_specific_execute_implementation(self, stmt: str, bind_params: SqlBindParameters) -> None:
        """Execute statement, returning nothing."""
        with self.get_connection() as connection:
            with connection.cursor() as cursor:
                logger.info(f"Executing SQL statment: {stmt}")
                cursor.execute(operation=stmt, parameters=self.params_or_none(bind_params))

    def _engine_specific_dry_run_implementation(self, stmt: str, bind_params: SqlBindParameters) -> None:
        """Check that query will run successfully without actually running the query, error if not."""
        stmt = f"EXPLAIN {stmt}"

        with self.get_connection() as connection:
            with connection.cursor() as cursor:
                logger.info(f"Executing SQL statment: {stmt}")
                cursor.execute(operation=stmt, parameters=self.params_or_none(bind_params))

                # If the plan contains errors, they won't be raised. Parse plan string to find & raise errors.
                result = str(cursor.fetchall_arrow()["plan"][0])
                if "org.apache.spark.sql.AnalysisException" in result:
                    error = result.split("== Physical Plan ==")[1].split(";")[0]
                    raise sql.exc.ServerOperationError(error)

    def create_table_from_dataframe(  # noqa: D
        self, sql_table: SqlTable, df: pd.DataFrame, chunk_size: Optional[int] = None
    ) -> None:
        logger.info(f"Creating table '{sql_table.sql}' from a DataFrame with {df.shape[0]} row(s)")
        start_time = time.time()
        with self.get_connection() as connection:
            with connection.cursor() as cursor:
                # Create table
                # update dtypes to convert None to NA in boolean columns.
                # This mirrors the SQLAlchemy schema detection logic in pandas.io.sql
                df = df.convert_dtypes()
                columns = df.columns
                columns_to_insert = []
                for i in range(len(df.columns)):
                    # Format as "column_name column_type"
                    columns_to_insert.append(f"{columns[i]} {PANDAS_TO_SQL_DTYPES[str(df[columns[i]].dtype).lower()]}")
                cursor.execute(f"CREATE TABLE IF NOT EXISTS {sql_table.sql} ({', '.join(columns_to_insert)})")

                # Insert rows
                values = ""
                for row in df.itertuples(index=False, name=None):
                    cells = ""
                    for cell in row:
                        cells += ", " if cells else ""
                        if pd.isnull(cell):
                            # Databricks does not support None, NA, nan, or NaT
                            cells += "null"
                        elif type(cell) in [str, pd.Timestamp]:
                            # Wrap cell in quotes & escape existing single quotes
                            escaped_cell = str(cell).replace("'", "\\'")
                            cells += f"'{escaped_cell}'"
                        else:
                            cells += str(cell)

                    values += (",\n" if values else "") + f"({cells})"
                    if chunk_size and len(values) == chunk_size:
                        cursor.execute(f"INSERT INTO {sql_table.sql} VALUES {values}")
                        values = ""
                if values:
                    cursor.execute(f"INSERT INTO {sql_table.sql} VALUES {values}")

        logger.info(f"Created table '{sql_table.sql}' from a DataFrame in {time.time() - start_time:.2f}s")

    def list_tables(self, schema_name: str) -> Sequence[str]:  # noqa: D
        with self.get_connection() as connection:
            with connection.cursor() as cursor:
                cursor.tables(schema_name=schema_name)
                return [table.TABLE_NAME for table in cursor.fetchall()]

    def cancel_submitted_queries(self) -> None:  # noqa: D
        pass

    def render_execution_param_key(self, execution_param_key: str) -> str:
        """Wrap execution parameter key with syntax accepted by engine."""
        return f"%({execution_param_key})s"



import json
from copy import deepcopy

from pathlib import Path
from typing import Dict, List, Union

from metricflow.model.parsing import schemas_internal

TOP_LEVEL_SCHEMAS = {
    "metric": "metric",
    "data_source": "data_source",
    "derived_group_by_element_schema": "derived_identifier",
    "materialization_schema": "materialization",
}

BASE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "title": "MetricFlow file schema",
}


def generate_explict_schema(schema_store: Dict) -> Dict:
    """Generates a single json schema object from the given schema store."""
    ref_to_definition_mapping = {key: f"#/definitions/{key}" for key in schema_store.keys()}
    definitions = {}
    for schema_name, _schema in schema_store.items():
        schema = deepcopy(_schema)

        rewritten_schema = _rewrite_refs(schema, ref_to_definition_mapping)
        assert isinstance(rewritten_schema, dict)

        if "definitions" in rewritten_schema:
            nested_definitions = rewritten_schema["definitions"]
            for name in nested_definitions.keys():
                definitions[name] = nested_definitions[name]
            rewritten_schema.pop("definitions", None)

        definitions[schema_name] = rewritten_schema

    properties = {}
    for schema_name, object_name in TOP_LEVEL_SCHEMAS.items():
        properties[object_name] = {"$ref": ref_to_definition_mapping[schema_name]}

    full_schema: Dict = deepcopy(BASE_SCHEMA)
    full_schema["properties"] = properties
    full_schema["definitions"] = definitions

    return full_schema


def _rewrite_refs(obj: Union[Dict, List, bool, str], mapping: Dict) -> Union[Dict, List, bool, str]:
    """Replaces the $refs from their names to their definition section identifiers."""
    if isinstance(obj, dict):
        _dict = {}
        for k, v in obj.items():
            if k == "$ref" and v in mapping:
                _dict[k] = mapping[v]
            else:
                _dict[k] = _rewrite_refs(v, mapping)
        return _dict
    if isinstance(obj, list):
        _list = []
        for element in obj:
            _list.append(_rewrite_refs(element, mapping))
        return _list
    return obj


def write_schema(schema: Dict, output_dir: str, file_name: str) -> None:
    """Writes the schema from the specified schema store to the given path"""

    path = Path(output_dir).resolve()
    path.mkdir(exist_ok=True)
    with open(path / file_name, "w") as f:
        json.dump(schema, f, indent=4, sort_keys=True)


if __name__ == "__main__":
    schema = generate_explict_schema(schemas_internal.schema_store)
    output_dir = str(Path(__file__).parent / "schemas")
    write_schema(schema, output_dir, "metricflow.json")



import pytest
import copy

from metricflow.model.model_validator import ModelValidator
from metricflow.model.validations.validator_helpers import ModelValidationException
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.unique_valid_name import MetricFlowReservedKeywords, UniqueAndValidNameRule
from metricflow.object_utils import flatten_nested_sequence
from metricflow.test.test_utils import find_data_source_with


def copied_model(simple_model__pre_transforms: UserConfiguredModel) -> UserConfiguredModel:  # noqa: D
    return copy.deepcopy(simple_model__pre_transforms)


"""
    Top Level Tests
    Top level elements include
    - Data Sources
    - Metrics
    - Metric Sets
    - Dimension Sets

    A name for any of these elements must be unique to all other top level elements
    except metrics. Metric names only need to be unique in comparison to other metric
    names.
"""


def test_duplicate_data_source_name(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    model = copied_model(simple_model__pre_transforms)
    duplicated_data_source = model.data_sources[0]
    model.data_sources.append(duplicated_data_source)
    with pytest.raises(
        ModelValidationException,
        match=rf"Can't use name `{duplicated_data_source.name}` for a data source when it was already used for a data source",
    ):
        ModelValidator().checked_validations(model)


def test_duplicate_metric_name(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa:D
    model = copied_model(simple_model__pre_transforms)
    duplicated_metric = model.metrics[0]
    model.metrics.append(duplicated_metric)
    with pytest.raises(
        ModelValidationException,
        match=rf"Can't use name `{duplicated_metric.name}` for a metric when it was already used for a metric",
    ):
        ModelValidator().checked_validations(model)


def test_duplicate_materalization_name(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa:D
    model = copied_model(simple_model__pre_transforms)
    duplicated_materialization = model.materializations[0]
    model.materializations.append(duplicated_materialization)
    with pytest.raises(
        ModelValidationException,
        match=rf"Can't use name `{duplicated_materialization.name}` for a materialization when it was already used for a materialization",
    ):
        ModelValidator().checked_validations(model)


def test_top_level_metric_can_have_same_name_as_any_other_top_level_item(
    simple_model__pre_transforms: UserConfiguredModel,
) -> None:  # noqa:D
    metric_name = simple_model__pre_transforms.metrics[0].name

    model_data_source = copied_model(simple_model__pre_transforms)
    model_materialization = copied_model(simple_model__pre_transforms)

    model_data_source.data_sources[0].name = metric_name
    model_materialization.materializations[0].name = model_data_source.metrics[0].name

    ModelValidator().checked_validations(model_data_source)
    ModelValidator().checked_validations(model_materialization)


def test_top_level_elements_cant_share_names_except_with_metrics(
    simple_model__pre_transforms: UserConfiguredModel,
) -> None:  # noqa:D
    data_source_name = simple_model__pre_transforms.data_sources[0].name
    model_ds_and_mat = copied_model(simple_model__pre_transforms)
    model_ds_and_mat.materializations[0].name = data_source_name

    with pytest.raises(
        ModelValidationException,
        match=rf"Can't use name `{data_source_name}` for a materialization when it was already used for a data source",
    ):
        ModelValidator().checked_validations(model_ds_and_mat)


"""
    Data Source Element Tests
    There are three types of data source elements
    - measures
    - identifiers
    - dimensions

    A name for any of these elements must be unique to all other element names
    for the given data source.
"""


def test_duplicate_measure_name(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa:D
    model = copied_model(simple_model__pre_transforms)

    # Ensure we have a usable data source for the test
    data_source_with_measures, _ = find_data_source_with(model, lambda data_source: len(data_source.measures) > 0)

    duplicated_measure = data_source_with_measures.measures[0]
    duplicated_measures_tuple = (data_source_with_measures.measures, (duplicated_measure,))
    data_source_with_measures.measures = flatten_nested_sequence(duplicated_measures_tuple)

    with pytest.raises(
        ModelValidationException,
        match=rf"can't use name `{duplicated_measure.reference.element_name}` for a measure when it was already used for a measure",
    ):
        ModelValidator().checked_validations(model)


def test_duplicate_dimension_name(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa:D
    model = copied_model(simple_model__pre_transforms)

    # Ensure we have a usable data source for the test
    data_source_with_dimensions, _ = find_data_source_with(model, lambda data_source: len(data_source.dimensions) > 0)

    duplicated_dimension = data_source_with_dimensions.dimensions[0]
    duplicated_dimensions_tuple = (data_source_with_dimensions.dimensions, (duplicated_dimension,))
    data_source_with_dimensions.dimensions = flatten_nested_sequence(duplicated_dimensions_tuple)

    with pytest.raises(
        ModelValidationException,
        match=rf"can't use name `{duplicated_dimension.reference.element_name}` for a "
        rf"dimension when it was already used for a dimension",
    ):
        ModelValidator().checked_validations(model)


def test_duplicate_identifier_name(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa:D
    model = copied_model(simple_model__pre_transforms)

    # Ensure we have a usable data source for the test
    data_source_with_identifiers, _ = find_data_source_with(model, lambda data_source: len(data_source.identifiers) > 0)

    duplicated_identifier = data_source_with_identifiers.identifiers[0]
    duplicated_identifiers_tuple = (data_source_with_identifiers.identifiers, (duplicated_identifier,))
    data_source_with_identifiers.identifiers = flatten_nested_sequence(duplicated_identifiers_tuple)

    with pytest.raises(
        ModelValidationException,
        match=rf"can't use name `{duplicated_identifier.reference.element_name}` for a identifier when it was already used for a identifier",
    ):
        ModelValidator().checked_validations(model)


"""
    Tests for valid naming
"""


def test_name_is_valid() -> None:  # noqa:D
    assert UniqueAndValidNameRule.check_valid_name("this_is_valid1") == []


def test_invalid_names() -> None:  # noqa:D
    # This is non-exhaustive, but covers some "common" error cases
    assert UniqueAndValidNameRule.check_valid_name("this is invalid") != []
    assert UniqueAndValidNameRule.check_valid_name("100%invalid") != []
    assert UniqueAndValidNameRule.check_valid_name("no-hyphens-allowed") != []
    assert UniqueAndValidNameRule.check_valid_name("punctuation.is.bad") != []
    assert UniqueAndValidNameRule.check_valid_name("cantwildstarme*") != []
    assert UniqueAndValidNameRule.check_valid_name("noemails@hellowdotcom") != []
    assert UniqueAndValidNameRule.check_valid_name("(no)(ordering)(operations)") != []
    assert UniqueAndValidNameRule.check_valid_name("path/to/invalid/name") != []
    assert UniqueAndValidNameRule.check_valid_name("#notwitterhere") != []
    assert UniqueAndValidNameRule.check_valid_name("_no_leading_underscore") != []
    assert UniqueAndValidNameRule.check_valid_name("no_trailing_underscore_") != []
    assert UniqueAndValidNameRule.check_valid_name("_definitely_no_leading_and_trailing_underscore_") != []

    # time granularity values are reserved
    assert UniqueAndValidNameRule.check_valid_name("day") != []
    assert UniqueAndValidNameRule.check_valid_name("week") != []
    assert UniqueAndValidNameRule.check_valid_name("month") != []
    assert UniqueAndValidNameRule.check_valid_name("quarter") != []
    assert UniqueAndValidNameRule.check_valid_name("year") != []


def test_reserved_name() -> None:  # noqa: D
    reserved_keyword = MetricFlowReservedKeywords.METRIC_TIME
    reserved_reason = MetricFlowReservedKeywords.get_reserved_reason(reserved_keyword)
    issues = UniqueAndValidNameRule.check_valid_name(reserved_keyword.value.lower())
    match = False
    for issue in issues:
        if issue.message.find(reserved_reason) != -1:
            match = True
    assert (
        match
    ), f"Did not find reason: '{reserved_reason}' in issues: {issues} for name: '{reserved_keyword.value.lower()}'"



from metricflow.instances import DataSourceReference, InstanceSet
from metricflow.plan_conversion.sql_dataset import SqlDataSet
from metricflow.sql.sql_plan import SqlSelectStatementNode


class DataSourceDataSet(SqlDataSet):
    """Similar to SqlDataSet, but contains metadata on the data source that was used to create this."""

    def __init__(  # noqa: D
        self,
        data_source_reference: DataSourceReference,
        instance_set: InstanceSet,
        sql_select_node: SqlSelectStatementNode,
    ) -> None:
        self._data_source_reference = data_source_reference
        super().__init__(instance_set=instance_set, sql_select_node=sql_select_node)

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}({self._data_source_reference})"

    @property
    def data_source_reference(self) -> DataSourceReference:  # noqa: D
        return self._data_source_reference



import pytest

from metricflow.dataflow.sql_column import SqlColumn
from metricflow.dataflow.sql_table import SqlTable
from metricflow.inference.context.data_warehouse import (
    ColumnProperties,
    DataWarehouseInferenceContext,
    InferenceColumnType,
    TableProperties,
)


@pytest.fixture
def warehouse_ctx() -> DataWarehouseInferenceContext:
    """A dummy DataWarehouseInferenceContext to be used as a fixture"""
    return DataWarehouseInferenceContext(
        table_props=[
            TableProperties(
                table=SqlTable.from_string("db.schema.table"),
                column_props=[
                    ColumnProperties(
                        column=SqlColumn.from_string("db.schema.table.id"),
                        type=InferenceColumnType.INTEGER,
                        row_count=1,
                        distinct_row_count=1,
                        is_nullable=True,
                        null_count=0,
                        min_value=0,
                        max_value=1,
                    ),
                    ColumnProperties(
                        column=SqlColumn.from_string("db.schema.table.othertable_id"),
                        type=InferenceColumnType.INTEGER,
                        row_count=2,
                        distinct_row_count=2,
                        is_nullable=True,
                        null_count=0,
                        min_value=0,
                        max_value=1,
                    ),
                    ColumnProperties(
                        column=SqlColumn.from_string("db.schema.table.test_column"),
                        type=InferenceColumnType.INTEGER,
                        row_count=1,
                        distinct_row_count=1,
                        is_nullable=True,
                        null_count=0,
                        min_value=0,
                        max_value=1,
                    ),
                    ColumnProperties(
                        column=SqlColumn.from_string("db.schema.othertable.othertable_id"),
                        type=InferenceColumnType.INTEGER,
                        row_count=2,
                        distinct_row_count=2,
                        is_nullable=True,
                        null_count=0,
                        min_value=0,
                        max_value=1,
                    ),
                ],
            )
        ]
    )



"""Classes required for defining metric definition object instances (see MdoInstance)."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, TypeVar, Generic, Tuple

from metricflow.aggregation_properties import AggregationState
from metricflow.column_assoc import ColumnAssociation
from metricflow.dataclass_serialization import SerializableDataclass
from metricflow.specs import (
    MeasureSpec,
    DimensionSpec,
    IdentifierSpec,
    MetricSpec,
    InstanceSpec,
    TimeDimensionSpec,
    InstanceSpecSet,
)


class ModelReference(SerializableDataclass):
    """A reference to something in the model.

    For example, a measure instance could have a defined_from field that has a model reference to the measure / data
    source that it is supposed to reference. Added for exploratory purposes, so whether this is needed is TBD.
    """

    pass


@dataclass(frozen=True)
class DataSourceReference(ModelReference):
    """A reference to a data source definition in the model."""

    data_source_name: str

    def __hash__(self) -> int:  # noqa: D
        return hash(self.data_source_name)


@dataclass(frozen=True)
class DataSourceElementReference(ModelReference):
    """A reference to an element definition in a data source definition in the model."""

    data_source_name: str
    element_name: str

    def is_from(self, ref: DataSourceReference) -> bool:
        """Returns true if this reference is from the same data source as the supplied reference."""
        return self.data_source_name == ref.data_source_name


@dataclass(frozen=True)
class MetricModelReference(ModelReference):
    """A reference to a metric definition in the model."""

    metric_name: str


@dataclass(frozen=True)
class MaterializationModelReference(ModelReference):
    """A reference to a materialization definition in the model."""

    materialization_name: str


# Type for the specification used in the instance.
SpecT = TypeVar("SpecT", bound=InstanceSpec)


# Type for the column association used in the instance.
ColumnNameAssociationT = TypeVar("ColumnNameAssociationT", bound=ColumnAssociation)


class MdoInstance(ABC, Generic[SpecT]):
    """An instance of a metric definition object.

    An instance is different from the metric definition object in that it correlates to columns in a data set and can be
    in different states. e.g. a measure instance can be aggregated, or a time dimension can be at a different
    granularity.
    """

    # The columns associated with this instance. Some instances may have multiple columns associated with it, e.g.
    # composite identifiers.
    associated_columns: Tuple[ColumnAssociation, ...]
    # The spec that describes this instance.
    spec: SpecT

    @property
    def associated_column(self) -> ColumnAssociation:
        """Helper for getting the associated column until support for multiple associated columns is added."""
        assert len(self.associated_columns) == 1
        return self.associated_columns[0]


# Instances for the major metric object types


@dataclass(frozen=True)
class DataSourceElementInstance(SerializableDataclass):  # noqa: D
    # This instance is derived from something defined in a data source.
    defined_from: Tuple[DataSourceElementReference, ...]


@dataclass(frozen=True)
class MeasureInstance(MdoInstance[MeasureSpec], DataSourceElementInstance):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: MeasureSpec
    aggregation_state: AggregationState


@dataclass(frozen=True)
class DimensionInstance(MdoInstance[DimensionSpec], DataSourceElementInstance):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: DimensionSpec


@dataclass(frozen=True)
class TimeDimensionInstance(MdoInstance[TimeDimensionSpec], DataSourceElementInstance):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: TimeDimensionSpec


@dataclass(frozen=True)
class IdentifierInstance(MdoInstance[IdentifierSpec], DataSourceElementInstance):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: IdentifierSpec


@dataclass(frozen=True)
class MetricInstance(MdoInstance[MetricSpec], SerializableDataclass):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: MetricSpec
    defined_from: Tuple[MetricModelReference, ...]


# Output type of transform function
TransformOutputT = TypeVar("TransformOutputT")


class InstanceSetTransform(Generic[TransformOutputT], ABC):
    """Function to use for transforming instance sets.

    This function interface should be used instead of manually modifying the members of an instance set so that it's
    easy to track, through "Find Usages", all transformations required for instance sets. This should make adding new
    fields (e.g. WeirdDimensionInstance) to instance sets easier to find and reason about.
    """

    @abstractmethod
    def transform(self, instance_set: InstanceSet) -> TransformOutputT:  # noqa: D
        pass


@dataclass(frozen=True)
class InstanceSet(SerializableDataclass):
    """A set that includes all instance types.

    Generally used to help represent that data that is flowing between nodes in the metric dataflow plan.
    """

    measure_instances: Tuple[MeasureInstance, ...] = ()
    dimension_instances: Tuple[DimensionInstance, ...] = ()
    time_dimension_instances: Tuple[TimeDimensionInstance, ...] = ()
    identifier_instances: Tuple[IdentifierInstance, ...] = ()
    metric_instances: Tuple[MetricInstance, ...] = ()

    def transform(self, transform_function: InstanceSetTransform[TransformOutputT]) -> TransformOutputT:  # noqa: D
        return transform_function.transform(self)

    @staticmethod
    def merge(instance_sets: List[InstanceSet]) -> InstanceSet:
        """Combine all instances from all instances into a single instance set.

        Instances will be de-duped based on their spec.
        """
        measure_instances: List[MeasureInstance] = []
        dimension_instances: List[DimensionInstance] = []
        time_dimension_instances: List[TimeDimensionInstance] = []
        identifier_instances: List[IdentifierInstance] = []
        metric_instances: List[MetricInstance] = []

        for instance_set in instance_sets:
            for measure_instance in instance_set.measure_instances:
                if measure_instance.spec not in {x.spec for x in measure_instances}:
                    measure_instances.append(measure_instance)
            for dimension_instance in instance_set.dimension_instances:
                if dimension_instance.spec not in {x.spec for x in dimension_instances}:
                    dimension_instances.append(dimension_instance)
            for time_dimension_instance in instance_set.time_dimension_instances:
                if time_dimension_instance.spec not in {x.spec for x in time_dimension_instances}:
                    time_dimension_instances.append(time_dimension_instance)
            for identifier_instance in instance_set.identifier_instances:
                if identifier_instance.spec not in {x.spec for x in identifier_instances}:
                    identifier_instances.append(identifier_instance)
            for metric_instance in instance_set.metric_instances:
                if metric_instance.spec not in {x.spec for x in metric_instances}:
                    metric_instances.append(metric_instance)

        return InstanceSet(
            measure_instances=tuple(measure_instances),
            dimension_instances=tuple(dimension_instances),
            time_dimension_instances=tuple(time_dimension_instances),
            identifier_instances=tuple(identifier_instances),
            metric_instances=tuple(metric_instances),
        )

    @property
    def spec_set(self) -> InstanceSpecSet:  # noqa: D
        return InstanceSpecSet(
            measure_specs=tuple(x.spec for x in self.measure_instances),
            dimension_specs=tuple(x.spec for x in self.dimension_instances),
            time_dimension_specs=tuple(x.spec for x in self.time_dimension_instances),
            identifier_specs=tuple(x.spec for x in self.identifier_instances),
            metric_specs=tuple(x.spec for x in self.metric_instances),
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar


class InferenceContext(ABC):
    """Encapsulates information that can be used by an inference signaling rule or inference policy."""

    pass


TContext = TypeVar("TContext", bound=InferenceContext)


class InferenceContextProvider(Generic[TContext], ABC):
    """Provides a populated inference context from some datasource."""

    @abstractmethod
    def get_context(self) -> TContext:
        """Fetch inference context data and return it."""
        pass



import logging
from typing import List
import pytest
from _pytest.fixtures import FixtureRequest

from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer, SqlQueryPlanRenderer
from metricflow.sql.sql_exprs import (
    SqlStringExpression,
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlComparisonExpression,
    SqlComparison,
    SqlFunctionExpression,
    SqlFunction,
)
from metricflow.sql.sql_plan import (
    SqlTableFromClauseNode,
    SqlSelectStatementNode,
    SqlSelectColumn,
    SqlJoinDescription,
    SqlOrderByDescription,
    SqlJoinType,
)
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.sql.compare_sql_plan import assert_rendered_sql_equal

logger = logging.getLogger(__name__)


@pytest.fixture
def default_sql_plan_renderer() -> SqlQueryPlanRenderer:  # noqa: D
    return DefaultSqlQueryPlanRenderer()


def test_component_rendering(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
) -> None:
    """Checks that all components of SELECT query are rendered for the 0, 1, >1 component count cases."""

    # Test single SELECT column
    select_columns = [
        SqlSelectColumn(
            expr=SqlFunctionExpression(sql_function=SqlFunction.SUM, sql_function_args=[SqlStringExpression("1")]),
            column_alias="bookings",
        ),
    ]

    from_source = SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="fct_bookings"))

    from_source = from_source
    from_source_alias = "a"
    joins_descs: List[SqlJoinDescription] = []
    where = None
    group_bys: List[SqlSelectColumn] = []
    order_bys: List[SqlOrderByDescription] = []

    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        select_node=SqlSelectStatementNode(
            description="test0",
            select_columns=tuple(select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            joins_descs=tuple(joins_descs),
            where=where,
            group_bys=tuple(group_bys),
            order_bys=tuple(order_bys),
        ),
        plan_id="plan0",
        sql_client=sql_client,
    )

    # Test multiple select column
    select_columns.extend(
        [
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(SqlColumnReference("b", "country")),
                column_alias="user__country",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(SqlColumnReference("c", "country")),
                column_alias="listing__country",
            ),
        ]
    )

    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        select_node=SqlSelectStatementNode(
            description="test1",
            select_columns=tuple(select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            joins_descs=tuple(joins_descs),
            where=where,
            group_bys=tuple(group_bys),
            order_bys=tuple(order_bys),
        ),
        plan_id="plan1",
        sql_client=sql_client,
    )

    # Test single join
    joins_descs.append(
        SqlJoinDescription(
            right_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="dim_users")),
            right_source_alias="b",
            on_condition=SqlComparisonExpression(
                left_expr=SqlColumnReferenceExpression(SqlColumnReference("a", "user_id")),
                comparison=SqlComparison.EQUALS,
                right_expr=SqlColumnReferenceExpression(SqlColumnReference("b", "user_id")),
            ),
            join_type=SqlJoinType.LEFT_OUTER,
        )
    )

    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        select_node=SqlSelectStatementNode(
            description="test2",
            select_columns=tuple(select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            joins_descs=tuple(joins_descs),
            where=where,
            group_bys=tuple(group_bys),
            order_bys=tuple(order_bys),
        ),
        plan_id="plan2",
        sql_client=sql_client,
    )

    # Test multiple join
    joins_descs.append(
        SqlJoinDescription(
            right_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="dim_listings")),
            right_source_alias="c",
            on_condition=SqlComparisonExpression(
                left_expr=SqlColumnReferenceExpression(SqlColumnReference("a", "user_id")),
                comparison=SqlComparison.EQUALS,
                right_expr=SqlColumnReferenceExpression(SqlColumnReference("c", "user_id")),
            ),
            join_type=SqlJoinType.LEFT_OUTER,
        )
    )

    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        select_node=SqlSelectStatementNode(
            description="test3",
            select_columns=tuple(select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            joins_descs=tuple(joins_descs),
            where=where,
            group_bys=tuple(group_bys),
            order_bys=tuple(order_bys),
        ),
        plan_id="plan3",
        sql_client=sql_client,
    )

    # Test single group by
    group_bys.append(
        SqlSelectColumn(
            expr=SqlColumnReferenceExpression(SqlColumnReference("b", "country")),
            column_alias="user__country",
        ),
    )

    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        select_node=SqlSelectStatementNode(
            description="test4",
            select_columns=tuple(select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            joins_descs=tuple(joins_descs),
            where=where,
            group_bys=tuple(group_bys),
            order_bys=tuple(order_bys),
        ),
        plan_id="plan4",
        sql_client=sql_client,
    )

    # Test multiple group bys
    group_bys.append(
        SqlSelectColumn(
            expr=SqlColumnReferenceExpression(SqlColumnReference("c", "country")),
            column_alias="listing__country",
        ),
    )

    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        select_node=SqlSelectStatementNode(
            description="test5",
            select_columns=tuple(select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            joins_descs=tuple(joins_descs),
            where=where,
            group_bys=tuple(group_bys),
            order_bys=tuple(order_bys),
        ),
        plan_id="plan5",
        sql_client=sql_client,
    )


def test_render_where(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
) -> None:
    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        select_node=SqlSelectStatementNode(
            description="test0",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="a", column_name="booking_value")
                    ),
                    column_alias="booking_value",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="fct_bookings")),
            from_source_alias="a",
            joins_descs=(),
            where=SqlComparisonExpression(
                left_expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="a", column_name="booking_value")
                ),
                comparison=SqlComparison.GREATER_THAN,
                right_expr=SqlStringExpression(
                    sql_expr="100",
                    requires_parenthesis=False,
                    used_columns=(),
                ),
            ),
            group_bys=(),
            order_bys=(),
        ),
        plan_id="plan0",
        sql_client=sql_client,
    )


def test_render_order_by(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
) -> None:
    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        select_node=SqlSelectStatementNode(
            description="test0",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="a", column_name="booking_value")
                    ),
                    column_alias="booking_value",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="a", column_name="bookings")
                    ),
                    column_alias="bookings",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="fct_bookings")),
            from_source_alias="a",
            joins_descs=(),
            where=None,
            group_bys=(),
            order_bys=(
                SqlOrderByDescription(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="a", column_name="booking_value")
                    ),
                    desc=False,
                ),
                SqlOrderByDescription(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="a", column_name="bookings")
                    ),
                    desc=True,
                ),
            ),
        ),
        plan_id="plan0",
        sql_client=sql_client,
    )


def test_render_limit(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
) -> None:
    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        select_node=SqlSelectStatementNode(
            description="test0",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="a", column_name="bookings")
                    ),
                    column_alias="bookings",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="fct_bookings")),
            from_source_alias="a",
            joins_descs=(),
            where=None,
            group_bys=(),
            order_bys=(),
            limit=1,
        ),
        plan_id="plan0",
        sql_client=sql_client,
    )



from collections import defaultdict
from typing import List, DefaultDict
from metricflow.instances import DataSourceReference

from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    DataSourceContext,
    DataSourceElementType,
    FileContext,
    ModelValidationRule,
    ValidationError,
    ValidationIssueType,
    validate_safely,
)


class ElementConsistencyRule(ModelValidationRule):
    """Checks that elements in data sources with the same name are of the same element type across the model

    This reduces the potential confusion that might arise from having an identifier named `country` and a dimension
    named `country` while allowing for things like the `user` identifier to exist in multiple data sources. Note not
    all element types allow duplicates, and there are separate validation rules for those cases. See, for example,
    the DataSourceMeasuresUniqueRule.
    """

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring model wide element consistency")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues = []
        element_name_to_types = ElementConsistencyRule._get_element_name_to_types(model=model)
        invalid_elements = {
            name: type_mapping for name, type_mapping in element_name_to_types.items() if len(type_mapping) > 1
        }

        for element_name, type_to_context in invalid_elements.items():
            # Sort these by value to ensure consistent error messaging
            types_used = [DataSourceElementType(v) for v in sorted(k.value for k in type_to_context.keys())]
            for element_type in types_used:
                data_source_contexts = type_to_context[element_type]
                data_source_names = {ctx.data_source.data_source_name for ctx in data_source_contexts}
                data_source_context = data_source_contexts[0]
                issues.append(
                    ValidationError(
                        context=data_source_context,
                        message=f"In data sources {data_source_names}, element `{element_name}` is of type "
                        f"{element_type}, but it is used as types {types_used} across the model.",
                    )
                )

        return issues

    @staticmethod
    def _get_element_name_to_types(
        model: UserConfiguredModel,
    ) -> DefaultDict[str, DefaultDict[DataSourceElementType, List[DataSourceContext]]]:
        """Create a mapping of all element names in the model to types with a list of associated DataSourceContexts"""
        element_types: DefaultDict[str, DefaultDict[DataSourceElementType, List[DataSourceContext]]] = defaultdict(
            lambda: defaultdict(list)
        )
        for data_source in model.data_sources:
            data_source_context = DataSourceContext(
                file_context=FileContext.from_metadata(metadata=data_source.metadata),
                data_source=DataSourceReference(data_source_name=data_source.name),
            )
            if data_source.measures:
                for measure in data_source.measures:
                    element_types[measure.name][DataSourceElementType.MEASURE].append(data_source_context)
            if data_source.dimensions:
                for dimension in data_source.dimensions:
                    element_types[dimension.name][DataSourceElementType.DIMENSION].append(data_source_context)
            if data_source.identifiers:
                for identifier in data_source.identifiers:
                    element_types[identifier.name][DataSourceElementType.IDENTIFIER].append(data_source_context)
        return element_types



from dataclasses import dataclass
from dbt.lib import get_dbt_config
from dbt import tracking
from dbt.parser.manifest import ManifestLoader as DbtManifestLoader, Manifest as DbtManifest
from metricflow.model.model_transformer import ModelTransformer
from metricflow.model.parsing.dir_to_model import ModelBuildResult
from metricflow.model.transformations.dbt_to_metricflow import DbtManifestTransformer
from typing import Optional


@dataclass
class DbtProfileArgs:
    """Class to represent dbt profile arguments

    dbt's get_dbt_config uses `getattr` to get values out of the passed in args.
    We cannot pass a simple dict, because `getattr` doesn't work for keys of a
    dictionary. Thus we create a simple object that `getattr` will work on.
    """

    profile: Optional[str] = None
    target: Optional[str] = None


def get_dbt_project_manifest(
    directory: str, profile: Optional[str] = None, target: Optional[str] = None
) -> DbtManifest:
    """Builds the dbt Manifest object from the dbt project"""

    profile_args = DbtProfileArgs(profile=profile, target=target)
    dbt_config = get_dbt_config(project_dir=directory, args=profile_args)
    # If we don't disable tracking, we have to setup a full
    # dbt User object to build the manifest
    tracking.disable_tracking()
    return DbtManifestLoader.get_full_manifest(config=dbt_config)


def parse_dbt_project_to_model(
    directory: str, profile: Optional[str] = None, target: Optional[str] = None
) -> ModelBuildResult:
    """Parse dbt model files in the given directory to a UserConfiguredModel."""
    manifest = get_dbt_project_manifest(directory=directory, profile=profile, target=target)
    build_result = DbtManifestTransformer(manifest=manifest).build_user_configured_model()
    transformed_model = ModelTransformer.pre_validation_transform_model(model=build_result.model)
    transformed_model = ModelTransformer.post_validation_transform_model(model=transformed_model)
    return ModelBuildResult(model=transformed_model, issues=build_result.issues)



import logging
from dataclasses import dataclass
from typing import Generic, Sequence, List, TypeVar, Optional, Set

from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.partitions import PartitionJoinResolver
from metricflow.dataflow.dataflow_plan import (
    ConstrainTimeRangeNode,
    BaseOutput,
    JoinToBaseOutputNode,
    FilterElementsNode,
    JoinDescription,
)
from metricflow.model.objects.elements.identifier import IdentifierType
from metricflow.model.semantics.semantic_containers import MAX_JOIN_HOPS
from metricflow.object_utils import pformat_big_objects
from metricflow.plan_conversion.sql_dataset import SqlDataSet
from metricflow.protocols.semantics import DataSourceSemanticsAccessor
from metricflow.references import TimeDimensionReference, IdentifierReference
from metricflow.spec_set_transforms import ToElementNameSet
from metricflow.specs import InstanceSpec, LinkableInstanceSpec, LinklessIdentifierSpec

SqlDataSetT = TypeVar("SqlDataSetT", bound=SqlDataSet)


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class MultiHopJoinCandidateLineage(Generic[SqlDataSetT]):
    """Describes how the multi-hop join candidate was formed.

    For example, if
    * bridge_source has the primary identifier account_id and the foreign identifier customer_id
    * customers_source has the primary identifier customer_id and dimension country
    * transactions_source has the transactions measure and the account_id foreign identifier

    Then the candidate lineage can be describe as
        first_node_to_join = bridge_source
        second_node_to_join = customers_source
    to get the country dimension.
    """

    first_node_to_join: BaseOutput[SqlDataSetT]
    second_node_to_join: BaseOutput[SqlDataSetT]
    join_second_node_by_identifier: LinklessIdentifierSpec


@dataclass(frozen=True)
class MultiHopJoinCandidate(Generic[SqlDataSetT]):
    """A candidate node containing linkable specs that is join of other nodes. It's used to resolve multi-hop queries.

    Also see MultiHopJoinCandidateLineage.
    """

    node_with_multi_hop_elements: BaseOutput[SqlDataSetT]
    lineage: MultiHopJoinCandidateLineage[SqlDataSetT]


class PreDimensionJoinNodeProcessor(Generic[SqlDataSetT]):
    """Processes source nodes before measures are joined to dimensions.

    Generally, the source nodes will be combined with other dataflow plan nodes to produce a new set of nodes to realize
    a condition of the query. For example, to realize a time range constraint, a ConstrainTimeRangeNode will be added
    to the source nodes.

    e.g.

    <SomeDataflowPlanNode/>

    ->

    <ConstrainTimeRangeNode>
        <SomeDataflowPlanNode/>
    </ConstrainTimeRangeNode>

    """

    def __init__(  # noqa: D
        self,
        data_source_semantics: DataSourceSemanticsAccessor,
        node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver[SqlDataSetT],
    ):
        self._node_data_set_resolver = node_data_set_resolver
        self._partition_resolver = PartitionJoinResolver(data_source_semantics)
        self._data_source_semantics = data_source_semantics

    def add_time_range_constraint(
        self,
        source_nodes: Sequence[BaseOutput[SqlDataSetT]],
        metric_time_dimension_reference: TimeDimensionReference,
        time_range_constraint: Optional[TimeRangeConstraint] = None,
    ) -> Sequence[BaseOutput[SqlDataSetT]]:
        """Adds a time range constraint node to the input nodes."""
        processed_nodes: List[BaseOutput[SqlDataSetT]] = []
        for source_node in source_nodes:

            # Constrain the time range if specified.
            if time_range_constraint:
                node_output_data_set = self._node_data_set_resolver.get_output_data_set(source_node)
                constrain_time = False
                for time_dimension_instance in node_output_data_set.instance_set.time_dimension_instances:
                    if (
                        time_dimension_instance.spec.reference == metric_time_dimension_reference
                        and len(time_dimension_instance.spec.identifier_links) == 0
                    ):
                        constrain_time = True
                        break
                if constrain_time:
                    processed_nodes.append(
                        ConstrainTimeRangeNode(parent_node=source_node, time_range_constraint=time_range_constraint)
                    )
                else:
                    processed_nodes.append(source_node)
            else:
                processed_nodes.append(source_node)
        return processed_nodes

    def _node_contains_identifier(
        self,
        node: BaseOutput[SqlDataSetT],
        identifier_reference: IdentifierReference,
        valid_types: Set[IdentifierType],
    ) -> bool:
        """Returns true if the output of the node contains an identifier of the given types."""
        data_set = self._node_data_set_resolver.get_output_data_set(node)

        for identifier_instance_in_first_node in data_set.instance_set.identifier_instances:
            identifier_spec_in_first_node = identifier_instance_in_first_node.spec

            if identifier_spec_in_first_node.reference != identifier_reference:
                continue

            if len(identifier_spec_in_first_node.identifier_links) > 0:
                continue

            assert (
                len(identifier_instance_in_first_node.defined_from) == 1
            ), "Multiple items in defined_from not yet supported"

            identifier = self._data_source_semantics.get_identifier_in_data_source(
                identifier_instance_in_first_node.defined_from[0]
            )
            if identifier is None:
                raise RuntimeError(
                    f"Invalid DataSourceElementReference {identifier_instance_in_first_node.defined_from[0]}"
                )

            if identifier.type in valid_types:
                return True

        return False

    def _get_candidates_nodes_for_multi_hop(
        self,
        desired_linkable_spec: LinkableInstanceSpec,
        nodes: Sequence[BaseOutput[SqlDataSetT]],
    ) -> Sequence[MultiHopJoinCandidate]:
        """Assemble nodes representing all possible one-hop joins"""

        if len(desired_linkable_spec.identifier_links) > MAX_JOIN_HOPS:
            raise NotImplementedError(
                f"Multi-hop joins with more than {MAX_JOIN_HOPS} identifier links not yet supported. "
                f"Got: {desired_linkable_spec}"
            )
        if len(desired_linkable_spec.identifier_links) != 2:
            return ()

        multi_hop_join_candidates: List[MultiHopJoinCandidate] = []
        logger.info(f"Creating nodes for {desired_linkable_spec}")

        for first_node_that_could_be_joined in nodes:
            data_set = self._node_data_set_resolver.get_output_data_set(first_node_that_could_be_joined)

            # Fan-out joins aren't supported, so the identifiers must be of these types to join.
            if not (
                self._node_contains_identifier(
                    node=first_node_that_could_be_joined,
                    identifier_reference=desired_linkable_spec.identifier_links[0],
                    valid_types={IdentifierType.PRIMARY, IdentifierType.UNIQUE},
                )
                and self._node_contains_identifier(
                    node=first_node_that_could_be_joined,
                    identifier_reference=desired_linkable_spec.identifier_links[1],
                    valid_types={IdentifierType.PRIMARY, IdentifierType.UNIQUE, IdentifierType.FOREIGN},
                )
            ):
                continue

            for second_node_that_could_be_joined in nodes:
                if not (
                    self._node_contains_identifier(
                        node=second_node_that_could_be_joined,
                        identifier_reference=desired_linkable_spec.identifier_links[1],
                        valid_types={IdentifierType.PRIMARY, IdentifierType.UNIQUE},
                    )
                ):
                    continue

                if second_node_that_could_be_joined.node_id == first_node_that_could_be_joined.node_id:
                    continue

                data_set_of_second_node_that_can_be_joined = self._node_data_set_resolver.get_output_data_set(
                    second_node_that_could_be_joined
                )

                # If the element name of the linkable spec doesn't exist in the joined data set, then it can't be
                # useful for obtaining that linkable spec.
                element_names_in_data_set = ToElementNameSet().transform(
                    data_set_of_second_node_that_can_be_joined.instance_set.spec_set
                )

                if desired_linkable_spec.element_name not in element_names_in_data_set:
                    continue

                # filter measures out of joinable_node
                specs = data_set_of_second_node_that_can_be_joined.instance_set.spec_set
                pass_specs = InstanceSpec.merge(
                    specs.dimension_specs + specs.identifier_specs + specs.time_dimension_specs
                )
                filtered_joinable_node = FilterElementsNode(second_node_that_could_be_joined, pass_specs)

                join_on_partition_dimensions = self._partition_resolver.resolve_partition_dimension_joins(
                    start_node_spec_set=data_set.instance_set.spec_set,
                    node_to_join_spec_set=data_set_of_second_node_that_can_be_joined.instance_set.spec_set,
                )
                join_on_partition_time_dimensions = self._partition_resolver.resolve_partition_time_dimension_joins(
                    start_node_spec_set=data_set.instance_set.spec_set,
                    node_to_join_spec_set=data_set_of_second_node_that_can_be_joined.instance_set.spec_set,
                )

                multi_hop_join_candidates.append(
                    MultiHopJoinCandidate(
                        node_with_multi_hop_elements=JoinToBaseOutputNode(
                            parent_node=first_node_that_could_be_joined,
                            join_targets=[
                                JoinDescription(
                                    join_node=filtered_joinable_node,
                                    join_on_identifier=LinklessIdentifierSpec.from_reference(
                                        desired_linkable_spec.identifier_links[1]
                                    ),
                                    join_on_partition_dimensions=join_on_partition_dimensions,
                                    join_on_partition_time_dimensions=join_on_partition_time_dimensions,
                                )
                            ],
                        ),
                        lineage=MultiHopJoinCandidateLineage(
                            first_node_to_join=first_node_that_could_be_joined,
                            second_node_to_join=second_node_that_could_be_joined,
                            # identifier_spec_in_first_node should already not have identifier links since we checked
                            # for that, but using this method for type checking.
                            join_second_node_by_identifier=LinklessIdentifierSpec.from_reference(
                                desired_linkable_spec.identifier_links[1]
                            ),
                        ),
                    )
                )

        for multi_hop_join_candidate in multi_hop_join_candidates:
            output_data_set = self._node_data_set_resolver.get_output_data_set(
                multi_hop_join_candidate.node_with_multi_hop_elements
            )
            logger.debug(
                f"Node {multi_hop_join_candidate.node_with_multi_hop_elements} has spec set:\n"
                f"{pformat_big_objects(output_data_set.instance_set.spec_set)}"
            )

        return multi_hop_join_candidates

    def add_multi_hop_joins(
        self, desired_linkable_specs: Sequence[LinkableInstanceSpec], nodes: Sequence[BaseOutput[SqlDataSetT]]
    ) -> Sequence[BaseOutput[SqlDataSetT]]:
        """Assemble nodes representing all possible one-hop joins"""

        all_multi_hop_join_candidates: List[MultiHopJoinCandidate[SqlDataSetT]] = []
        lineage_for_all_multi_hop_join_candidates: Set[MultiHopJoinCandidateLineage[SqlDataSetT]] = set()

        for desired_linkable_spec in desired_linkable_specs:
            for multi_hop_join_candidate in self._get_candidates_nodes_for_multi_hop(
                desired_linkable_spec=desired_linkable_spec,
                nodes=nodes,
            ):
                # Dedupe candidates that are the same join.
                if multi_hop_join_candidate.lineage not in lineage_for_all_multi_hop_join_candidates:
                    all_multi_hop_join_candidates.append(multi_hop_join_candidate)
                    lineage_for_all_multi_hop_join_candidates.add(multi_hop_join_candidate.lineage)

        return list(x.node_with_multi_hop_elements for x in all_multi_hop_join_candidates) + list(nodes)

    def remove_unnecessary_nodes(
        self,
        desired_linkable_specs: Sequence[LinkableInstanceSpec],
        nodes: Sequence[BaseOutput[SqlDataSetT]],
        metric_time_dimension_reference: TimeDimensionReference,
    ) -> Sequence[BaseOutput[SqlDataSetT]]:
        """Filters out many of the nodes that can't possibly be useful for joins to obtain the desired linkable specs.

        A simple filter is to remove any nodes that don't share a common element with the query. Having a common element
        doesn't mean that the node will be useful, but not having common elements definitely means it's not useful.
        """
        relevant_element_names = {x.element_name for x in desired_linkable_specs}.union(
            {y.element_name for x in desired_linkable_specs for y in x.identifier_links}
        )

        # The metric time dimension is used everywhere, so don't count it unless specifically desired in linkable spec
        # that has identifier links.
        metric_time_dimension_used_in_linked_spec = any(
            [
                len(linkable_spec.identifier_links) > 0
                and linkable_spec.element_name == metric_time_dimension_reference.element_name
                for linkable_spec in desired_linkable_specs
            ]
        )

        if (
            metric_time_dimension_reference.element_name in relevant_element_names
            and not metric_time_dimension_used_in_linked_spec
        ):
            relevant_element_names.remove(metric_time_dimension_reference.element_name)

        logger.info(f"Relevant names are: {relevant_element_names}")

        relevant_nodes = []

        for node in nodes:
            data_set = self._node_data_set_resolver.get_output_data_set(node)
            element_names_in_data_set = ToElementNameSet().transform(data_set.instance_set.spec_set)

            if len(element_names_in_data_set.intersection(relevant_element_names)) > 0:
                relevant_nodes.append(node)

        return relevant_nodes



from __future__ import annotations

from collections import OrderedDict
from typing import Dict, Any

from metricflow.model.objects.base import HashableBaseModel


class SqlBindParameters(HashableBaseModel):
    """Helps to build execution parameters during SQL query rendering.

    These can be used as per https://docs.sqlalchemy.org/en/14/core/tutorial.html#using-textual-sql
    """

    param_dict: Dict[str, str] = OrderedDict()

    def update(self, additional_params: SqlBindParameters) -> None:
        """Add the parameters to this set, mutating it."""
        for key, value in additional_params.param_dict.items():
            if key in self.param_dict and self.param_dict[key] != value:
                raise RuntimeError(
                    f"Conflict with key {key} in merging parameters. "
                    f"Existing params: {self.param_dict} Additional params: {additional_params}"
                )
        self.param_dict.update(additional_params.param_dict)

    def __eq__(self, other: Any) -> bool:  # type: ignore  # noqa: D
        return isinstance(other, SqlBindParameters) and self.param_dict == other.param_dict



import logging
from typing import List, Optional

from metricflow.sql.optimizer.sql_query_plan_optimizer import SqlQueryPlanOptimizer
from metricflow.sql.sql_exprs import SqlColumnReferenceExpression, SqlColumnReference
from metricflow.sql.sql_plan import (
    SqlQueryPlanNode,
    SqlQueryPlanNodeVisitor,
    SqlSelectQueryFromClauseNode,
    SqlTableFromClauseNode,
    SqlSelectStatementNode,
    SqlOrderByDescription,
    SqlJoinDescription,
)

logger = logging.getLogger(__name__)


class SqlSubQueryReducerVisitor(SqlQueryPlanNodeVisitor[SqlQueryPlanNode]):
    """Visits the SQL query plan to simplify sub-queries. On each visit, return a simplfied node"""

    def _reduce_parents(
        self,
        node: SqlSelectStatementNode,
    ) -> SqlSelectStatementNode:
        """Apply the reducing operation to the parent select statements."""
        return SqlSelectStatementNode(
            description=node.description,
            select_columns=node.select_columns,
            from_source=node.from_source.accept(self),
            from_source_alias=node.from_source_alias,
            joins_descs=tuple(
                SqlJoinDescription(
                    right_source=x.right_source.accept(self),
                    right_source_alias=x.right_source_alias,
                    on_condition=x.on_condition,
                    join_type=x.join_type,
                )
                for x in node.join_descs
            ),
            group_bys=node.group_bys,
            order_bys=node.order_bys,
            where=node.where,
            limit=node.limit,
        )

    def _reduce_is_possible(self, node: SqlSelectStatementNode) -> bool:  # noqa: D
        """Returns true if the given node can be reduced with the parent node.

        Reducing this node means eliminating the SELECT of this node and merging it with the parent SELECT. This
        checks for obvious cases where reducing can't happen, but there are cases where reducing is possible, but this
        returns false for ease of reasoning.
        """

        # If this node has multiple parents (i.e. a join), then this can't be collapsed.
        is_join = len(node.join_descs) > 0
        has_multiple_parent_nodes = len(node.parent_nodes) > 1
        if has_multiple_parent_nodes or is_join:
            return False

        assert len(node.parent_nodes) == 1
        parent_node = node.parent_nodes[0]

        # If the parent node is not a select statement, then this can't be collapsed. e.g. with a table as a parent like
        # SELECT foo FROM bar
        if not parent_node.as_select_node:
            return False
        parent_select_node = parent_node.as_select_node

        # More conditions where we don't want to collapse. It's not impossible with these cases, but not reducing in
        # these cases for simplicity.

        # Reducing a where is tricky as it requires the expressions to be re-written.
        if node.where:
            return False

        # Group bys are hard to reduce.
        if len(node.group_bys) > 0:
            return False

        # Similar with order bys.
        if len(node.order_bys) > 0 and len(parent_select_node.order_bys) > 0:
            return False

        # Can't reduce if the sub-query selects more columns. This generally won't be hit as long as the column pruner
        # runs before this optimization.
        if len(parent_select_node.select_columns) > len(node.select_columns):
            return False

        # For now, simplify only in cases where the select expressions directly reference columns in the parent query
        # and the alias is the same as the column reference. This means that the column references don't have to be
        # re-written when this is collapsed into the parent subquery.
        for select_column in node.select_columns:
            if not select_column.expr.as_column_reference_expression:
                return False
            column_reference_expression = select_column.expr.as_column_reference_expression
            assert column_reference_expression
            if column_reference_expression.col_ref.column_name != select_column.column_alias:
                return False

        # Same thing with order by
        for order_by in node.order_bys:
            if not order_by.expr.as_column_reference_expression:
                return False

        return True

    @staticmethod
    def _find_matching_table_alias(node: SqlSelectStatementNode, column_alias: str) -> Optional[str]:
        for select_column in node.select_columns:
            if select_column.column_alias == column_alias:
                column_reference_expr = select_column.expr.as_column_reference_expression
                if column_reference_expr:
                    return column_reference_expr.col_ref.table_alias
        return None

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> SqlQueryPlanNode:  # noqa: D
        node_with_reduced_parents = self._reduce_parents(node)

        if not self._reduce_is_possible(node_with_reduced_parents):
            return node_with_reduced_parents

        assert len(node_with_reduced_parents.parent_nodes) == 1
        parent_node = node_with_reduced_parents.parent_nodes[0]
        parent_select_node = parent_node.as_select_node
        assert parent_select_node

        # At this point, the query should look similar to
        #
        #     SELECT b.foo, b.baz
        #     FROM (
        #       SELECT a.foo, a.baz FROM bar a
        #       LIMIT 10
        #     ) b
        #     ORDER by b.baz
        #     LIMIT 1

        # The order by in the parent doesn't matter since the order by in this node will "overwrite" the order in the
        # parent as long as the parent has no limits.
        new_order_by: List[SqlOrderByDescription] = []
        if node_with_reduced_parents.order_bys:
            for order_by_item in node_with_reduced_parents.order_bys:
                order_by_item_expr = order_by_item.expr.as_column_reference_expression
                assert order_by_item_expr

                # Re-write the order by as a column expression that references the table alias
                table_alias_in_parent = SqlSubQueryReducerVisitor._find_matching_table_alias(
                    parent_select_node, order_by_item_expr.col_ref.column_name
                )
                if not table_alias_in_parent:
                    return node_with_reduced_parents
                new_order_by.append(
                    SqlOrderByDescription(
                        expr=SqlColumnReferenceExpression(
                            SqlColumnReference(
                                table_alias=table_alias_in_parent,
                                column_name=order_by_item_expr.col_ref.column_name,
                            )
                        ),
                        desc=order_by_item.desc,
                    )
                )

        # The limit should be the min of this SELECT limit and the parent SELECT limit.
        new_limit: Optional[int] = node_with_reduced_parents.limit
        if new_limit is None:
            new_limit = parent_select_node.limit
        elif parent_select_node.limit is not None:
            new_limit = min(new_limit, parent_select_node.limit)

        return SqlSelectStatementNode(
            description="\n".join([parent_select_node.description, node_with_reduced_parents.description]),
            select_columns=parent_select_node.select_columns,
            from_source=parent_select_node.from_source,
            from_source_alias=parent_select_node.from_source_alias,
            joins_descs=parent_select_node.join_descs,
            group_bys=parent_select_node.group_bys,
            order_bys=tuple(new_order_by),
            where=parent_select_node.where,
            limit=new_limit,
        )

    def visit_table_from_clause_node(self, node: SqlTableFromClauseNode) -> SqlQueryPlanNode:  # noqa: D
        return node

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> SqlQueryPlanNode:  # noqa: D
        return node


class SqlSubQueryReducer(SqlQueryPlanOptimizer):
    """Simplify queries by eliminating sub-queries when possible.

    e.g. from

    SELECT b.foo
    FROM (
      SELECT a.foo FROM bar a
    ) b

    to

    SELECT a.foo FROM bar a
    """

    def optimize(self, node: SqlQueryPlanNode) -> SqlQueryPlanNode:  # noqa: D
        return node.accept(SqlSubQueryReducerVisitor())



import pytest
from _pytest.fixtures import FixtureRequest

from metricflow.dataflow.sql_table import SqlTable
from metricflow.sql.optimizer.rewriting_sub_query_reducer import SqlRewritingSubQueryReducer
from metricflow.sql.sql_exprs import (
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlComparisonExpression,
    SqlComparison,
    SqlStringLiteralExpression,
    SqlFunctionExpression,
    SqlFunction,
    SqlStringExpression,
)
from metricflow.sql.sql_plan import (
    SqlSelectStatementNode,
    SqlSelectColumn,
    SqlTableFromClauseNode,
    SqlOrderByDescription,
    SqlJoinDescription,
    SqlJoinType,
)
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.sql.compare_sql_plan import assert_default_rendered_sql_equal


@pytest.fixture
def base_select_statement() -> SqlSelectStatementNode:
    """SELECT statement used to build test cases.

    -- src3
    SELECT
      SUM(src2.col0) AS bookings
      src2.col1 AS ds
      -- src2
      SELECT
        src1.bookings AS bookings
        , src1.ds AS ds
      FROM (
        -- src1
        SELECT
          1 AS bookings
          , src0.ds AS ds
        FROM demo.fct_bookings src0
        LIMIT 2
      ) src1
      WHERE src1.ds >= "2020-01-01"
      LIMIT 1
    ) src2
    WHERE src2.ds <= "2020-01-05"
    GROUP BY src2.ds
    ORDER BY src2.ds
    """
    return SqlSelectStatementNode(
        description="src3",
        select_columns=(
            SqlSelectColumn(
                expr=SqlFunctionExpression(
                    sql_function=SqlFunction.SUM,
                    sql_function_args=[
                        SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(table_alias="src2", column_name="bookings")
                        )
                    ],
                ),
                column_alias="bookings",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(col_ref=SqlColumnReference(table_alias="src2", column_name="ds")),
                column_alias="ds",
            ),
        ),
        from_source=SqlSelectStatementNode(
            description="src2",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="src1", column_name="bookings")
                    ),
                    column_alias="bookings",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(col_ref=SqlColumnReference(table_alias="src1", column_name="ds")),
                    column_alias="ds",
                ),
            ),
            from_source=SqlSelectStatementNode(
                description="src1",
                select_columns=(
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(
                                table_alias="src0",
                                column_name="bookings",
                            )
                        ),
                        column_alias="bookings",
                    ),
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(
                                table_alias="src0",
                                column_name="ds",
                            )
                        ),
                        column_alias="ds",
                    ),
                ),
                from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="fct_bookings")),
                from_source_alias="src0",
                joins_descs=(),
                group_bys=(),
                order_bys=(),
                limit=2,
            ),
            from_source_alias="src1",
            joins_descs=(),
            where=SqlComparisonExpression(
                left_expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias="src1",
                        column_name="ds",
                    )
                ),
                comparison=SqlComparison.GREATER_THAN_OR_EQUALS,
                right_expr=SqlStringLiteralExpression("2020-01-01"),
            ),
            group_bys=(),
            order_bys=(),
            limit=1,
        ),
        from_source_alias="src2",
        joins_descs=(),
        group_bys=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias="src2",
                        column_name="ds",
                    )
                ),
                column_alias="ds",
            ),
        ),
        where=SqlComparisonExpression(
            left_expr=SqlColumnReferenceExpression(
                SqlColumnReference(
                    table_alias="src2",
                    column_name="ds",
                )
            ),
            comparison=SqlComparison.LESS_THAN_OR_EQUALS,
            right_expr=SqlStringLiteralExpression("2020-01-05"),
        ),
        order_bys=(
            SqlOrderByDescription(
                expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias="src2",
                        column_name="ds",
                    )
                ),
                desc=False,
            ),
        ),
        limit=None,
    )


def test_reduce_sub_query(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    base_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests a case where an outer query should be reduced into its inner query with merged LIMIT expressions"""
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=base_select_statement,
        plan_id="before_reducing",
    )

    sub_query_reducer = SqlRewritingSubQueryReducer()

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=sub_query_reducer.optimize(base_select_statement),
        plan_id="after_reducing",
    )


@pytest.fixture
def join_select_statement() -> SqlSelectStatementNode:
    """SELECT statement with a join used to build test cases.

    SELECT
      SUM(bookings_src.bookings) AS bookings
      listings_src.country_latest AS listing__country_latest
      bookings_src.ds AS ds
    FROM (
      SELECT
        fct_bookings_src.booking AS bookings
        , 1 AS ds
        , fct_bookings_src.listing_id AS listing
      FROM demo.fct_bookings fct_bookings_src
      WHERE fct_bookings_src.ds >= "2020-01-01"
    ) bookings_src
    JOIN (
      SELECT
        dim_listings_src.country_latest AS country_latest
        , dim_listings_src.listing_id AS listing
      FROM demo.dim_listings dim_listings_src
    ) listings_src
    ON bookings_src.listing = listings_src.listing
    GROUP BY bookings_src.ds
    """
    return SqlSelectStatementNode(
        description="query",
        select_columns=(
            SqlSelectColumn(
                expr=SqlFunctionExpression(
                    sql_function=SqlFunction.SUM,
                    sql_function_args=[
                        SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(table_alias="bookings_src", column_name="bookings")
                        )
                    ],
                ),
                column_alias="bookings",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="listings_src", column_name="country_latest")
                ),
                column_alias="listing__country_latest",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="bookings_src", column_name="ds")
                ),
                column_alias="ds",
            ),
        ),
        from_source=SqlSelectStatementNode(
            description="bookings_src",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="fct_bookings_src", column_name="booking")
                    ),
                    column_alias="bookings",
                ),
                SqlSelectColumn(
                    expr=SqlStringExpression(sql_expr="1", requires_parenthesis=False, used_columns=()),
                    column_alias="ds",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="fct_bookings_src", column_name="listing_id")
                    ),
                    column_alias="listing",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="fct_bookings")),
            from_source_alias="fct_bookings_src",
            joins_descs=(),
            where=None,
            group_bys=(),
            order_bys=(),
            limit=None,
        ),
        from_source_alias="bookings_src",
        joins_descs=(
            SqlJoinDescription(
                right_source=SqlSelectStatementNode(
                    description="listings_src",
                    select_columns=(
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(table_alias="dim_listings_src", column_name="country")
                            ),
                            column_alias="country_latest",
                        ),
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(table_alias="dim_listings_src", column_name="listing_id")
                            ),
                            column_alias="listing",
                        ),
                    ),
                    from_source=SqlTableFromClauseNode(
                        sql_table=SqlTable(schema_name="demo", table_name="dim_listings")
                    ),
                    from_source_alias="dim_listings_src",
                    joins_descs=(),
                    where=None,
                    group_bys=(),
                    order_bys=(),
                    limit=None,
                ),
                right_source_alias="listings_src",
                on_condition=SqlComparisonExpression(
                    left_expr=SqlColumnReferenceExpression(
                        SqlColumnReference(table_alias="bookings_src", column_name="listing"),
                    ),
                    comparison=SqlComparison.EQUALS,
                    right_expr=SqlColumnReferenceExpression(
                        SqlColumnReference(table_alias="listings_src", column_name="listing"),
                    ),
                ),
                join_type=SqlJoinType.LEFT_OUTER,
            ),
        ),
        group_bys=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias="bookings_src",
                        column_name="ds",
                    )
                ),
                column_alias="ds",
            ),
        ),
        where=SqlComparisonExpression(
            left_expr=SqlColumnReferenceExpression(
                SqlColumnReference(
                    table_alias="bookings_src",
                    column_name="ds",
                )
            ),
            comparison=SqlComparison.LESS_THAN_OR_EQUALS,
            right_expr=SqlStringLiteralExpression("2020-01-05"),
        ),
        order_bys=(),
        limit=None,
    )


def test_reduce_join(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    join_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests a case where reducing occurs on a JOIN"""
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=join_select_statement,
        plan_id="before_reducing",
    )

    sub_query_reducer = SqlRewritingSubQueryReducer()

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=sub_query_reducer.optimize(join_select_statement),
        plan_id="after_reducing",
    )


@pytest.fixture
def colliding_select_statement() -> SqlSelectStatementNode:
    """SELECT statement that would cause a collision of table aliases if reduced.

    SELECT
      SUM(bookings_src.bookings) AS bookings
      listings_src.country_latest AS listing__country_latest
      bookings_src.ds AS ds
    FROM (
      SELECT
        colliding_alias.booking AS bookings
        , colliding_alias.ds AS ds
        , colliding_alias.listing_id AS listing
      FROM demo.fct_bookings colliding_alias
      WHERE fct_bookings_src.ds >= "2020-01-01"
    ) bookings_src
    JOIN (
      SELECT
        colliding_alias.country_latest AS country_latest
        , colliding_alias.listing_id AS listing
      FROM demo.dim_listings colliding_alias
    ) listings_src
    ON bookings_src.listing = listings_src.listing
    GROUP BY bookings_src.ds
    """
    return SqlSelectStatementNode(
        description="query",
        select_columns=(
            SqlSelectColumn(
                expr=SqlFunctionExpression(
                    sql_function=SqlFunction.SUM,
                    sql_function_args=[
                        SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(table_alias="bookings_src", column_name="bookings")
                        )
                    ],
                ),
                column_alias="bookings",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="listings_src", column_name="listing__country_latest")
                ),
                column_alias="listing__country_latest",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="bookings_src", column_name="ds")
                ),
                column_alias="ds",
            ),
        ),
        from_source=SqlSelectStatementNode(
            description="bookings_src",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="colliding_alias", column_name="booking")
                    ),
                    column_alias="bookings",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="colliding_alias", column_name="ds")
                    ),
                    column_alias="ds",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="colliding_alias", column_name="listing_id")
                    ),
                    column_alias="listing",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="fct_bookings")),
            from_source_alias="colliding_alias",
            joins_descs=(),
            where=None,
            group_bys=(),
            order_bys=(),
            limit=None,
        ),
        from_source_alias="bookings_src",
        joins_descs=(
            SqlJoinDescription(
                right_source=SqlSelectStatementNode(
                    description="listings_src",
                    select_columns=(
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(table_alias="colliding_alias", column_name="country")
                            ),
                            column_alias="country_latest",
                        ),
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(table_alias="colliding_alias", column_name="listing_id")
                            ),
                            column_alias="listing",
                        ),
                    ),
                    from_source=SqlTableFromClauseNode(
                        sql_table=SqlTable(schema_name="demo", table_name="dim_listings")
                    ),
                    from_source_alias="colliding_alias",
                    joins_descs=(),
                    where=None,
                    group_bys=(),
                    order_bys=(),
                    limit=None,
                ),
                right_source_alias="listings_src",
                on_condition=SqlComparisonExpression(
                    left_expr=SqlColumnReferenceExpression(
                        SqlColumnReference(table_alias="bookings_src", column_name="listing"),
                    ),
                    comparison=SqlComparison.EQUALS,
                    right_expr=SqlColumnReferenceExpression(
                        SqlColumnReference(table_alias="listings_src", column_name="listing"),
                    ),
                ),
                join_type=SqlJoinType.LEFT_OUTER,
            ),
        ),
        group_bys=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias="bookings_src",
                        column_name="ds",
                    )
                ),
                column_alias="ds",
            ),
        ),
        where=SqlComparisonExpression(
            left_expr=SqlColumnReferenceExpression(
                SqlColumnReference(
                    table_alias="bookings_src",
                    column_name="ds",
                )
            ),
            comparison=SqlComparison.LESS_THAN_OR_EQUALS,
            right_expr=SqlStringLiteralExpression("2020-01-05"),
        ),
        order_bys=(),
        limit=None,
    )


def test_colliding_alias(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    colliding_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests a case where reducing occurs on a JOIN"""
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=colliding_select_statement,
        plan_id="before_reducing",
    )

    sub_query_reducer = SqlRewritingSubQueryReducer()

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=sub_query_reducer.optimize(colliding_select_statement),
        plan_id="after_reducing",
    )


@pytest.fixture
def reduce_all_join_select_statement() -> SqlSelectStatementNode:
    """SELECT statement with a join where all sources can be reduced.

    SELECT
      SUM(bookings_src.bookings) AS bookings
      listings_src1.country_latest AS listing__country_latest
      listings_src2.capacity_latest AS listing__capacity_latest
      bookings_src.ds AS ds
    FROM (
      SELECT
        fct_bookings_src.booking AS bookings
        , fct_bookings_src.ds AS ds
        , fct_bookings_src.listing_id AS listing
      FROM demo.fct_bookings fct_bookings_src
      WHERE fct_bookings_src.ds >= "2020-01-01"
    ) bookings_src
    JOIN (
      SELECT
        dim_listings_src1.country_latest AS country_latest
        , dim_listings_src1.listing_id AS listing
      FROM demo.dim_listings dim_listings_src1
    ) listings_src1
    ON bookings_src.listing = listings_src.listing
    JOIN (
      SELECT
        dim_listings_src2.country_latest AS capacity_latest
        , dim_listings_src2.listing_id AS listing
      FROM demo.dim_listings dim_listings_src2
    ) listings_src2
    ON listing_src1.listing = listings_src2.listing
    GROUP BY bookings_src.ds, listings_src1.country_latest, listings_src2.capacity_latest
    """
    return SqlSelectStatementNode(
        description="query",
        select_columns=(
            SqlSelectColumn(
                expr=SqlFunctionExpression(
                    sql_function=SqlFunction.SUM,
                    sql_function_args=[
                        SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(table_alias="bookings_src", column_name="bookings")
                        )
                    ],
                ),
                column_alias="bookings",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="listings_src1", column_name="country_latest")
                ),
                column_alias="listing__country_latest",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="listings_src2", column_name="capacity_latest")
                ),
                column_alias="listing__capacity_latest",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="bookings_src", column_name="ds")
                ),
                column_alias="ds",
            ),
        ),
        from_source=SqlSelectStatementNode(
            description="bookings_src",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="fct_bookings_src", column_name="booking")
                    ),
                    column_alias="bookings",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="fct_bookings_src", column_name="ds")
                    ),
                    column_alias="ds",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="fct_bookings_src", column_name="listing_id")
                    ),
                    column_alias="listing",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="fct_bookings")),
            from_source_alias="fct_bookings_src",
            joins_descs=(),
            where=None,
            group_bys=(),
            order_bys=(),
            limit=None,
        ),
        from_source_alias="bookings_src",
        joins_descs=(
            SqlJoinDescription(
                right_source=SqlSelectStatementNode(
                    description="listings_src1",
                    select_columns=(
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(table_alias="dim_listings_src1", column_name="country")
                            ),
                            column_alias="country_latest",
                        ),
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(table_alias="dim_listings_src1", column_name="listing_id")
                            ),
                            column_alias="listing",
                        ),
                    ),
                    from_source=SqlTableFromClauseNode(
                        sql_table=SqlTable(schema_name="demo", table_name="dim_listings")
                    ),
                    from_source_alias="dim_listings_src1",
                    joins_descs=(),
                    where=None,
                    group_bys=(),
                    order_bys=(),
                    limit=None,
                ),
                right_source_alias="listings_src1",
                on_condition=SqlComparisonExpression(
                    left_expr=SqlColumnReferenceExpression(
                        SqlColumnReference(table_alias="bookings_src", column_name="listing"),
                    ),
                    comparison=SqlComparison.EQUALS,
                    right_expr=SqlColumnReferenceExpression(
                        SqlColumnReference(table_alias="listings_src1", column_name="listing"),
                    ),
                ),
                join_type=SqlJoinType.LEFT_OUTER,
            ),
            SqlJoinDescription(
                right_source=SqlSelectStatementNode(
                    description="listings_src2",
                    select_columns=(
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(table_alias="dim_listings_src2", column_name="capacity")
                            ),
                            column_alias="capacity_latest",
                        ),
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(table_alias="dim_listings_src2", column_name="listing_id")
                            ),
                            column_alias="listing",
                        ),
                    ),
                    from_source=SqlTableFromClauseNode(
                        sql_table=SqlTable(schema_name="demo", table_name="dim_listings")
                    ),
                    from_source_alias="dim_listings_src2",
                    joins_descs=(),
                    where=None,
                    group_bys=(),
                    order_bys=(),
                    limit=None,
                ),
                right_source_alias="listings_src2",
                on_condition=SqlComparisonExpression(
                    left_expr=SqlColumnReferenceExpression(
                        SqlColumnReference(table_alias="listings_src1", column_name="listing"),
                    ),
                    comparison=SqlComparison.EQUALS,
                    right_expr=SqlColumnReferenceExpression(
                        SqlColumnReference(table_alias="listings_src2", column_name="listing"),
                    ),
                ),
                join_type=SqlJoinType.LEFT_OUTER,
            ),
        ),
        group_bys=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias="bookings_src",
                        column_name="ds",
                    )
                ),
                column_alias="ds",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias="listings_src1",
                        column_name="country_latest",
                    )
                ),
                column_alias="listing__country_latest",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias="listings_src2",
                        column_name="capacity_latest",
                    )
                ),
                column_alias="listing__capacity_latest",
            ),
        ),
        where=None,
        order_bys=(),
        limit=None,
    )


def test_reduce_all_join_sources(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    reduce_all_join_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests a case where reducing occurs all all sources on a JOIN"""
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=reduce_all_join_select_statement,
        plan_id="before_reducing",
    )

    sub_query_reducer = SqlRewritingSubQueryReducer()

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=sub_query_reducer.optimize(reduce_all_join_select_statement),
        plan_id="after_reducing",
    )



import logging
import textwrap

from metricflow.object_utils import pretty_format, pformat_big_objects
from metricflow.specs import DimensionSpec, IdentifierReference

logger = logging.getLogger(__name__)


def test_pretty_print() -> None:  # noqa: D
    assert pretty_format("foo") == "foo"
    assert pretty_format(1) == "1"
    assert pretty_format(0.1) == "0.1"
    assert pretty_format(["bar", "baz"]) == "['bar', 'baz']"
    assert pretty_format(("bar", "baz")) == "('bar', 'baz')"
    assert pretty_format({"foo": 1}) == "{'foo': 1}"


def test_pformat_big_objects() -> None:  # noqa: D
    dimension_spec = DimensionSpec(
        element_name="country_latest", identifier_links=(IdentifierReference(element_name="listing"),)
    )

    assert pformat_big_objects(dimension_spec) == (
        textwrap.dedent(
            """\
            {'class': 'DimensionSpec',
             'element_name': 'country_latest',
             'identifier_links': ({'class': 'IdentifierReference',
                                   'element_name': 'listing'},)}
            """
        ).rstrip()
    )
    logger.error(f"Output:\n{pformat_big_objects(dimension_spec=dimension_spec)}")
    assert pformat_big_objects(dimension_spec=dimension_spec) == (
        textwrap.dedent(
            """\
            dimension_spec:
                {'class': 'DimensionSpec',
                 'element_name': 'country_latest',
                 'identifier_links': ({'class': 'IdentifierReference',
                                       'element_name': 'listing'},)}
            """
        ).rstrip()
    )



import datetime

import pytest

from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.source_node import SourceNodeBuilder
from metricflow.dataset.convert_data_source import DataSourceToDataSetConverter
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.dataset.dataset import DataSet
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.references import MetricReference
from metricflow.test.time.metric_time_dimension import MTD_SPEC_DAY, MTD_SPEC_MONTH, MTD_SPEC_YEAR, MTD_REFERENCE
from metricflow.time.time_granularity import TimeGranularity
from metricflow.time.time_granularity_solver import (
    TimeGranularitySolver,
    PartialTimeDimensionSpec,
    RequestTimeGranularityException,
)


@pytest.fixture(scope="session")
def time_granularity_solver(  # noqa: D
    extended_date_semantic_model: SemanticModel,
    time_spine_source: TimeSpineSource,
) -> TimeGranularitySolver:
    column_association_resolver = DefaultColumnAssociationResolver(extended_date_semantic_model)
    node_output_resolver = DataflowPlanNodeOutputDataSetResolver[DataSourceDataSet](
        column_association_resolver=column_association_resolver,
        semantic_model=extended_date_semantic_model,
        time_spine_source=time_spine_source,
    )
    to_data_set_converter = DataSourceToDataSetConverter(column_association_resolver)
    source_data_sets = [
        to_data_set_converter.create_sql_source_data_set(x)
        for x in extended_date_semantic_model.user_configured_model.data_sources
    ]

    source_node_builder = SourceNodeBuilder(extended_date_semantic_model)
    source_nodes = source_node_builder.create_from_data_sets(source_data_sets)
    return TimeGranularitySolver(
        semantic_model=extended_date_semantic_model,
        source_nodes=source_nodes,
        node_output_resolver=node_output_resolver,
    )


def test_validate_day_granuarity_for_day_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    time_granularity_solver.validate_time_granularity(
        metric_references=[MetricReference(element_name="bookings")],
        time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.DAY)],
    )


def test_validate_month_granuarity_for_day_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    time_granularity_solver.validate_time_granularity(
        metric_references=[MetricReference(element_name="bookings")],
        time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.MONTH)],
    )


def test_validate_month_granuarity_for_month_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    time_granularity_solver.validate_time_granularity(
        metric_references=[MetricReference(element_name="bookings_monthly")],
        time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.MONTH)],
    )


def test_validate_month_granuarity_for_day_and_month_metrics(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    time_granularity_solver.validate_time_granularity(
        metric_references=[MetricReference(element_name="bookings"), MetricReference(element_name="bookings_monthly")],
        time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.MONTH)],
    )


def test_validate_year_granularity_for_day_and_month_metrics(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    time_granularity_solver.validate_time_granularity(
        metric_references=[MetricReference(element_name="bookings"), MetricReference(element_name="bookings_monthly")],
        time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.YEAR)],
    )


def test_validate_day_granuarity_for_month_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    with pytest.raises(RequestTimeGranularityException):
        time_granularity_solver.validate_time_granularity(
            metric_references=[MetricReference(element_name="bookings_monthly")],
            time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.DAY)],
        )


def test_validate_day_granularity_for_day_and_month_metric(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    with pytest.raises(RequestTimeGranularityException):
        time_granularity_solver.validate_time_granularity(
            metric_references=[
                MetricReference(element_name="bookings"),
                MetricReference(element_name="bookings_monthly"),
            ],
            time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.DAY)],
        )


PARTIAL_PTD_SPEC = PartialTimeDimensionSpec(element_name=DataSet.metric_time_dimension_name(), identifier_links=())


def test_granularity_solution_for_day_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_references=[MetricReference(element_name="bookings")],
        partial_time_dimension_specs=[PARTIAL_PTD_SPEC],
        metric_time_dimension_reference=MTD_REFERENCE,
    ) == {
        PARTIAL_PTD_SPEC: MTD_SPEC_DAY,
    }


def test_granularity_solution_for_month_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_references=[MetricReference(element_name="bookings_monthly")],
        partial_time_dimension_specs=[PARTIAL_PTD_SPEC],
        metric_time_dimension_reference=MTD_REFERENCE,
    ) == {
        PARTIAL_PTD_SPEC: MTD_SPEC_MONTH,
    }


def test_granularity_solution_for_day_and_month_metrics(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_references=[MetricReference(element_name="bookings"), MetricReference(element_name="bookings_monthly")],
        partial_time_dimension_specs=[PARTIAL_PTD_SPEC],
        metric_time_dimension_reference=MTD_REFERENCE,
    ) == {PARTIAL_PTD_SPEC: MTD_SPEC_MONTH}


def test_time_granularity_parameter(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_references=[MetricReference(element_name="bookings"), MetricReference(element_name="bookings_monthly")],
        partial_time_dimension_specs=[PARTIAL_PTD_SPEC],
        metric_time_dimension_reference=MTD_REFERENCE,
        time_granularity=TimeGranularity.YEAR,
    ) == {PARTIAL_PTD_SPEC: MTD_SPEC_YEAR}


def test_invalid_time_granularity_parameter(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    with pytest.raises(RequestTimeGranularityException, match="Can't use time granularity.*"):
        time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
            metric_references=[
                MetricReference(element_name="bookings"),
                MetricReference(element_name="bookings_monthly"),
            ],
            partial_time_dimension_specs=[PARTIAL_PTD_SPEC],
            metric_time_dimension_reference=MTD_REFERENCE,
            time_granularity=TimeGranularity.DAY,
        )


def test_adjusted_time_constraint(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 1),
            end_time=datetime.datetime(2020, 1, 15),
        ),
        time_granularity=TimeGranularity.DAY,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 1),
        end_time=datetime.datetime(2020, 1, 15),
    )

    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 1),
            end_time=datetime.datetime(2020, 1, 15),
        ),
        time_granularity=TimeGranularity.MONTH,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 1),
        end_time=datetime.datetime(2020, 1, 31),
    )

    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 15),
            end_time=datetime.datetime(2020, 2, 15),
        ),
        time_granularity=TimeGranularity.MONTH,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 1),
        end_time=datetime.datetime(2020, 2, 29),
    )


def test_adjusted_time_constraint_for_week_granularity(
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    """Tests the time range adjustment behavior for weekly granularities"""
    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 8),
            end_time=datetime.datetime(2020, 1, 15),
        ),
        time_granularity=TimeGranularity.WEEK,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 6),
        end_time=datetime.datetime(2020, 1, 19),
    )


def test_adjusted_time_constraint_at_boundaries(
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    """Tests the time range adjustment behavior at the min / max boundaries"""
    assert (
        time_granularity_solver.adjust_time_range_to_granularity(
            time_range_constraint=TimeRangeConstraint(
                start_time=TimeRangeConstraint.all_time().start_time + datetime.timedelta(days=-30),
                end_time=TimeRangeConstraint.all_time().end_time + datetime.timedelta(days=30),
            ),
            time_granularity=TimeGranularity.WEEK,
        )
        == TimeRangeConstraint.all_time()
    )



from __future__ import annotations

import pytest

from metricflow.dataflow.builder.costing import DefaultCostFunction
from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.time_spine import TimeSpineSource, TimeSpineTableBuilder
from metricflow.protocols.sql_client import SqlClient
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.fixtures.sql_client_fixtures import sql_client  # noqa: F401, F403

"""
Using 'function' scope to make ID generation more deterministic for the dataflow plan builder fixtures..

Using 'session' scope can result in other 'session' scope fixtures causing ID consistency issues.
"""


@pytest.fixture
def composite_dataflow_plan_builder(  # noqa: D
    composite_identifier_semantic_model: SemanticModel,
    consistent_id_object_repository: ConsistentIdObjectRepository,
    time_spine_source: TimeSpineSource,
) -> DataflowPlanBuilder[DataSourceDataSet]:

    return DataflowPlanBuilder(
        source_nodes=consistent_id_object_repository.composite_model_source_nodes,
        semantic_model=composite_identifier_semantic_model,
        cost_function=DefaultCostFunction[DataSourceDataSet](),
        time_spine_source=time_spine_source,
    )


@pytest.fixture
def dataflow_plan_builder(  # noqa: D
    simple_semantic_model: SemanticModel,
    consistent_id_object_repository: ConsistentIdObjectRepository,
    time_spine_source: TimeSpineSource,
) -> DataflowPlanBuilder[DataSourceDataSet]:

    return DataflowPlanBuilder(
        source_nodes=consistent_id_object_repository.simple_model_source_nodes,
        semantic_model=simple_semantic_model,
        cost_function=DefaultCostFunction[DataSourceDataSet](),
        time_spine_source=time_spine_source,
    )


@pytest.fixture
def multihop_dataflow_plan_builder(  # noqa: D
    multi_hop_join_semantic_model: SemanticModel,
    consistent_id_object_repository: ConsistentIdObjectRepository,
    time_spine_source: TimeSpineSource,
) -> DataflowPlanBuilder[DataSourceDataSet]:

    return DataflowPlanBuilder(
        source_nodes=consistent_id_object_repository.multihop_model_source_nodes,
        semantic_model=multi_hop_join_semantic_model,
        cost_function=DefaultCostFunction[DataSourceDataSet](),
        time_spine_source=time_spine_source,
    )


@pytest.fixture(scope="session")
def time_spine_source(  # noqa: D
    sql_client: SqlClient, mf_test_session_state: MetricFlowTestSessionState  # noqa: F811
) -> TimeSpineSource:
    time_spine_source = TimeSpineSource(schema_name=mf_test_session_state.mf_system_schema)
    time_spine_table_builder = TimeSpineTableBuilder(time_spine_source=time_spine_source, sql_client=sql_client)
    time_spine_table_builder.create_if_necessary()
    return time_spine_source



from __future__ import annotations

import logging

from metricflow.instances import (
    InstanceSet,
)
from metricflow.model.validations.unique_valid_name import MetricFlowReservedKeywords
from metricflow.references import TimeDimensionReference
from metricflow.specs import TimeDimensionSpec
from metricflow.time.time_granularity import TimeGranularity


logger = logging.getLogger(__name__)


class DataSet:
    """Describes a set of data that a source node in the dataflow plan contains."""

    def __init__(self, instance_set: InstanceSet) -> None:  # noqa:
        self._instance_set = instance_set

    @property
    def instance_set(self) -> InstanceSet:
        """Returns the instances contained in this dataset."""
        return self._instance_set

    @staticmethod
    def metric_time_dimension_reference() -> TimeDimensionReference:
        """Returns a special reference that means 'the aggregation time dimension for all measures in the data set'

        Or to put in another way, if metrics for those measures were plotted together on a graph, this is the name of
        the time dimension for the x-axis.
        """
        return TimeDimensionReference(element_name=MetricFlowReservedKeywords.METRIC_TIME.value)

    @staticmethod
    def metric_time_dimension_name() -> str:
        """See metric_time_dimension_reference(), but in name form for use in queries."""
        return DataSet.metric_time_dimension_reference().element_name

    @staticmethod
    def metric_time_dimension_spec(time_granularity: TimeGranularity) -> TimeDimensionSpec:
        """Spec that corresponds to DataSet.metric_time_dimension_reference"""
        return TimeDimensionSpec(
            element_name=DataSet.metric_time_dimension_reference().element_name,
            identifier_links=(),
            time_granularity=time_granularity,
        )

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}()"



from __future__ import annotations

import logging
import textwrap
from abc import ABC
from collections import namedtuple
from dataclasses import dataclass
from typing import List

import jinja2

from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql.sql_exprs import (
    SqlExpressionNodeVisitor,
    SqlColumnReferenceExpression,
    SqlStringExpression,
    SqlComparisonExpression,
    SqlExpressionNode,
    SqlFunction,
    SqlFunctionExpression,
    SqlNullExpression,
    SqlLogicalExpression,
    SqlStringLiteralExpression,
    SqlIsNullExpression,
    SqlCastToTimestampExpression,
    SqlDateTruncExpression,
    SqlTimeDeltaExpression,
    SqlRatioComputationExpression,
    SqlColumnAliasReferenceExpression,
    SqlBetweenExpression,
)
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SqlExpressionRenderResult:
    """The result of rendering an SQL expression tree to a string."""

    sql: str
    execution_parameters: SqlBindParameters


class SqlExpressionRenderer(SqlExpressionNodeVisitor[SqlExpressionRenderResult], ABC):
    """Renders SqlExpressions into strings"""

    def render_sql_expr(self, sql_expr: SqlExpressionNode) -> SqlExpressionRenderResult:
        """Render the given expression to a string."""
        return sql_expr.accept(self)

    @property
    def double_data_type(self) -> str:
        """Property for the double data type, for engine-specific type casting

        TODO: Eliminate this in favor of some kind of engine properties struct
        """
        return "DOUBLE"


class DefaultSqlExpressionRenderer(SqlExpressionRenderer):
    """Renders the SQL query plan assuming ANSI SQL."""

    def visit_string_expr(self, node: SqlStringExpression) -> SqlExpressionRenderResult:  # noqa: D
        """Renders an arbitrary string expression like 1+1=2"""
        return SqlExpressionRenderResult(sql=node.sql_expr, execution_parameters=node.execution_parameters)

    def visit_column_reference_expr(self, node: SqlColumnReferenceExpression) -> SqlExpressionRenderResult:  # noqa: D
        """Render a reference to a column in a table like my_table.my_col"""
        return SqlExpressionRenderResult(
            sql=(
                f"{node.col_ref.table_alias}.{node.col_ref.column_name}"
                if node.should_render_table_alias
                else node.col_ref.column_name
            ),
            execution_parameters=SqlBindParameters(),
        )

    def visit_column_alias_reference_expr(  # noqa: D
        self, node: SqlColumnAliasReferenceExpression
    ) -> SqlExpressionRenderResult:
        """Render a reference to a column without a known table alias. e.g. foo.bar vs bar."""
        return SqlExpressionRenderResult(
            sql=node.column_alias,
            execution_parameters=SqlBindParameters(),
        )

    def visit_comparison_expr(self, node: SqlComparisonExpression) -> SqlExpressionRenderResult:
        """Render a comparison expression like 1 = 2"""
        combined_params = SqlBindParameters()

        left_expr_rendered = self.render_sql_expr(node.left_expr)
        combined_params.update(left_expr_rendered.execution_parameters)

        right_expr_rendered = self.render_sql_expr(node.right_expr)
        combined_params.update(right_expr_rendered.execution_parameters)

        # To avoid issues with operator precedence, use parenthesis to group the left / right expressions if they
        # contain operators.
        return SqlExpressionRenderResult(
            # Render a + b = c
            sql=(
                (f"({left_expr_rendered.sql})" if node.left_expr.requires_parenthesis else left_expr_rendered.sql)
                + f" {node.comparison.value} "
                + (f"({right_expr_rendered.sql})" if node.right_expr.requires_parenthesis else right_expr_rendered.sql)
            ),
            execution_parameters=combined_params,
        )

    def visit_function_expr(self, node: SqlFunctionExpression) -> SqlExpressionRenderResult:  # noqa: D
        """Render a function call like CONCAT(a, b)"""
        args_rendered = [self.render_sql_expr(x) for x in node.sql_function_args]
        combined_params = SqlBindParameters()
        for arg_rendered in args_rendered:
            combined_params.update(arg_rendered.execution_parameters)

        distinct_prefix = "DISTINCT " if SqlFunction.is_distinct_aggregation(node.sql_function) else ""
        args_string = ", ".join([x.sql for x in args_rendered])

        return SqlExpressionRenderResult(
            sql=f"{node.sql_function.value}({distinct_prefix}{args_string})",
            execution_parameters=combined_params,
        )

    def visit_null_expr(self, node: SqlNullExpression) -> SqlExpressionRenderResult:  # noqa: D
        return SqlExpressionRenderResult(
            sql="NULL",
            execution_parameters=SqlBindParameters(),
        )

    def visit_string_literal_expr(self, node: SqlStringLiteralExpression) -> SqlExpressionRenderResult:  # noqa: D
        return SqlExpressionRenderResult(
            sql=f"'{node.literal_value}'",
            execution_parameters=SqlBindParameters(),
        )

    def visit_logical_expr(self, node: SqlLogicalExpression) -> SqlExpressionRenderResult:  # noqa: D
        RenderedExpr = namedtuple("RenderedExpr", ["expr", "requires_parenthesis"])
        args_rendered = [RenderedExpr(self.render_sql_expr(x), x.requires_parenthesis) for x in node.args]
        combined_parameters = SqlBindParameters()
        args_sql: List[str] = []

        can_be_rendered_in_one_line = sum(len(x.expr.sql) for x in args_rendered) < 60

        for arg_rendered in args_rendered:
            combined_parameters.update(arg_rendered.expr.execution_parameters)
            arg_sql = self._render_logical_arg(
                arg_rendered.expr, arg_rendered.requires_parenthesis, render_in_one_line=can_be_rendered_in_one_line
            )
            args_sql.append(arg_sql)

        sql = f" {node.operator.value} ".join(args_sql)

        return SqlExpressionRenderResult(
            sql=sql,
            execution_parameters=combined_parameters,
        )

    @staticmethod
    def _render_logical_arg(
        arg_rendered: SqlExpressionRenderResult, requires_parenthesis: bool, render_in_one_line: bool
    ) -> str:
        # Put everything on 1 line for short expressions, but otherwise put one expression per line
        # Note: multi-line expressions are always enclosed in parentheses.

        # e.g.
        # Put everything on 1 line for short expressions like:
        #
        # (1 < 2) AND foo
        #
        # but for long expressions do:
        #
        #  (
        #    some_long_expression1
        #  ) AND (
        #    some_long_expression2
        #  ) AND (
        #    some_long_expression3
        # )

        if render_in_one_line:
            return arg_rendered.sql if not requires_parenthesis else f"({arg_rendered.sql})"
        else:
            return (
                jinja2.Template(
                    textwrap.dedent(
                        """\
                    (
                      {{ arg_sql | indent(2) }}
                    )
                    """
                    )
                )
                .render(arg_sql=arg_rendered.sql)
                .rstrip()
            )

    def visit_is_null_expr(self, node: SqlIsNullExpression) -> SqlExpressionRenderResult:  # noqa: D
        arg_rendered = self.render_sql_expr(node.arg)

        return SqlExpressionRenderResult(
            sql=f"{arg_rendered.sql} IS NULL" if not node.arg.requires_parenthesis else f"({arg_rendered.sql}) IS NULL",
            execution_parameters=arg_rendered.execution_parameters,
        )

    def visit_cast_to_timestamp_expr(self, node: SqlCastToTimestampExpression) -> SqlExpressionRenderResult:  # noqa: D
        arg_rendered = self.render_sql_expr(node.arg)
        return SqlExpressionRenderResult(
            sql=f"CAST({arg_rendered.sql} AS TIMESTAMP)",
            execution_parameters=arg_rendered.execution_parameters,
        )

    def visit_date_trunc_expr(self, node: SqlDateTruncExpression) -> SqlExpressionRenderResult:  # noqa: D
        arg_rendered = self.render_sql_expr(node.arg)

        return SqlExpressionRenderResult(
            sql=f"DATE_TRUNC('{node.time_granularity.value}', {arg_rendered.sql})",
            execution_parameters=arg_rendered.execution_parameters,
        )

    def visit_time_delta_expr(self, node: SqlTimeDeltaExpression) -> SqlExpressionRenderResult:  # noqa: D
        arg_rendered = node.arg.accept(self)
        if node.grain_to_date:
            return SqlExpressionRenderResult(
                sql=f"DATE_TRUNC('{node.granularity.value}', {arg_rendered.sql}::timestamp)",
                execution_parameters=arg_rendered.execution_parameters,
            )

        count = node.count
        granularity = node.granularity
        if granularity == TimeGranularity.QUARTER:
            granularity = TimeGranularity.MONTH
            count *= 3
        return SqlExpressionRenderResult(
            sql=f"DATEADD({granularity.value}, -{count}, {arg_rendered.sql})",
            execution_parameters=arg_rendered.execution_parameters,
        )

    def visit_ratio_computation_expr(self, node: SqlRatioComputationExpression) -> SqlExpressionRenderResult:
        """Render the ratio computation for a ratio metric

        This requires both a type cast to a floating point type (default to DOUBLE, engine-permitting) and
        the requisite division between numerator and denominator
        """
        rendered_numerator = self.render_sql_expr(node.numerator)
        rendered_denominator = self.render_sql_expr(node.denominator)

        numerator_sql = f"CAST({rendered_numerator.sql} AS {self.double_data_type})"
        denominator_sql = f"CAST(NULLIF({rendered_denominator.sql}, 0) AS {self.double_data_type})"

        execution_parameters = SqlBindParameters()
        execution_parameters.update(rendered_numerator.execution_parameters)
        execution_parameters.update(rendered_denominator.execution_parameters)

        return SqlExpressionRenderResult(
            sql=f"{numerator_sql} / {denominator_sql}",
            execution_parameters=execution_parameters,
        )

    def visit_between_expr(self, node: SqlBetweenExpression) -> SqlExpressionRenderResult:  # noqa: D
        rendered_column_arg = self.render_sql_expr(node.column_arg)
        rendered_start_expr = self.render_sql_expr(node.start_expr)
        rendered_end_expr = self.render_sql_expr(node.end_expr)

        execution_parameters = SqlBindParameters()
        execution_parameters.update(rendered_column_arg.execution_parameters)
        execution_parameters.update(rendered_start_expr.execution_parameters)
        execution_parameters.update(rendered_end_expr.execution_parameters)

        return SqlExpressionRenderResult(
            sql=f"{rendered_column_arg.sql} BETWEEN {rendered_start_expr.sql} AND {rendered_end_expr.sql}",
            execution_parameters=execution_parameters,
        )



from __future__ import annotations

import json
import os
from abc import ABC, abstractmethod
from typing import Any, Callable, ClassVar, Generic, Iterator, TypeVar

from pydantic import BaseModel, root_validator

from metricflow.errors.errors import ParsingException
from metricflow.model.parsing.yaml_loader import ParsingContext, PARSING_CONTEXT_KEY

# Type alias for the implicit "Any" type used as input and output for Pydantic's parsing API
PydanticParseableValueType = Any  # type: ignore[misc]


class HashableBaseModel(BaseModel):
    """Extends BaseModel with a generic hash function"""

    def __hash__(self) -> int:  # noqa: D
        return hash(json.dumps(self.json(sort_keys=True), sort_keys=True))


class FrozenBaseModel(HashableBaseModel):
    """Similar to HashableBaseModel but faux immutable."""

    class Config:
        """Pydantic feature."""

        allow_mutation = False

    def to_pretty_json(self) -> str:
        """Convert to a pretty JSON representation."""
        raw_json_str = self.json()
        json_obj = json.loads(raw_json_str)
        return json.dumps(json_obj, indent=4)

    def __str__(self) -> str:  # noqa: D
        return self.__repr__()


class ModelWithMetadataParsing(BaseModel):
    """Pydantic model object with a root validator for converting ParsingContext into Metadata]

    To use this validator the model class in question MUST have a field with the following base
    specification:

      metadata: Optional[Metadata]

    This class does NOT define the metadata itself because Pydantic's parsing and validation can be
    dependent on the order in which the model properties are listed in the class definition.
    Therefore, we prefer to keep this generic and allow callers to add the metadata field wherever
    we encounter parsing context information.
    """

    __METADATA_KEY__: ClassVar[str] = "metadata"

    @root_validator(pre=True)
    @classmethod
    def extract_metadata_from_parsing_context(cls, values: PydanticParseableValueType) -> PydanticParseableValueType:
        """Takes info from parsing context and converts it to a Metadata model object

        Per Pydantic's processing logic, this runs on the collection of input data for whatever model
        object is doing its parser walk. Since we set pre to True this should happen before any of the
        properties in the model is allowed to access the input values, which allows us to update them
        to include the appropriate inputs for defined Metadata object properties.
        """
        if not isinstance(values, dict):
            raise ValueError(
                f"Input values should be an object (dict) type, but got type({values}) with value: {values}"
            )

        # yes, this is confusing, but values is a dict, so it must have keys
        keys = values.keys()
        if cls.__METADATA_KEY__ in keys:
            # Sometimes people pass metadata in directly, e.g., in measure proxy metrics. Let Pydantic handle it.
            return values

        if PARSING_CONTEXT_KEY not in keys:
            # TODO: determine whether or not we want measure metadata tagged to measure proxy metrics. If we do,
            # add enforcement and make Metadata a non-optional element wherever it is set
            return values

        context = values.pop(PARSING_CONTEXT_KEY)
        if not isinstance(context, ParsingContext):
            raise ParsingException(
                f"Parsing context should always be a ParsingContext object, but we got a {type(context)} "
                f"with value: {context} inside payload: {values}"
            )

        values[cls.__METADATA_KEY__] = {
            "repo_file_path": context.filename,
            "file_slice": {
                "filename": os.path.split(context.filename)[-1],
                "content": context.content,
                "start_line_number": context.start_line,
                "end_line_number": context.end_line,
            },
        }
        return values


ModelObjectT_co = TypeVar("ModelObjectT_co", covariant=True, bound=BaseModel)


class PydanticCustomInputParser(ABC, Generic[ModelObjectT_co]):
    """Implements required methods for enabling custom parsing for Pydantic BaseModel objects

    This abstract class helper is for the specific case where model object classes need to do custom parsing
    prior to object initialization, meaning that the inputs to the initializer itself must be parsed in a
    manner custom to the object, but without updating the input values for the rest of the model hierarchy.

    This class will NOT allow for custom handling of dict or object type inputs, assuming that they should be
    processed and validated in the standard Pydantic way. Any subclass needing to mutate the inputs from a dict
    or object type in a subclass-specific way should still be able to do so via the standard Pydantic approach
    of adding a method decorated with @validator or @root_validator, which will work internally to initialization
    and validation of that model object itself.
    """

    @classmethod
    def __get_validators__(cls) -> Iterator[Callable]:
        """Pydantic magic method for allowing parsing of arbitrary input on parse_obj invocation

        This allow for parsing and validation prior to object initialization. Most classes implementing this
        interface in our model are doing so because the input value from user-supplied YAML will be a string
        representation rather than the structured object type.
        """
        yield cls.__parse_with_custom_handling

    @classmethod
    def __parse_with_custom_handling(cls: ModelObjectT_co, input: PydanticParseableValueType) -> ModelObjectT_co:
        """Core method for handling common valid - or easily validated - input types

        Pydantic objects can commonly appear as JSON object types (from, e.g., deserializing a Pydantic-serialized
        model) or direct instances of the model object class (from, e.g., initializing an object and passing it in
        to the initializer of a containing model object). This internal wrapper handles both of these cases in the
        standard Pydantic way of either validating the object type input on initialization, or passing the
        already-validated instance along as the output.

        Note this will also pass any Pydantic instance initialized via the `construct` method, which means it is
        possible for a caller to thread an unvalidated - and therefore improperly initialized - instance through.
        However, this is part of the contract with `construct` - it is only to be used when the inputs are known
        to the caller to be pre-validated, and so we do not bother guarding against that here.
        """
        if isinstance(input, dict):
            return cls(**input)
        elif isinstance(input, cls):
            return input
        else:
            return cls._from_yaml_value(input)

    @classmethod
    @abstractmethod
    def _from_yaml_value(cls: ModelObjectT_co, input: PydanticParseableValueType) -> ModelObjectT_co:
        """Abstract method for providing object-specific parsing logic"""
        raise NotImplementedError()



from metricflow.sql.render.expr_renderer import (
    DefaultSqlExpressionRenderer,
    SqlExpressionRenderer,
    SqlExpressionRenderResult,
)
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer
from metricflow.sql.sql_exprs import (
    SqlCastToTimestampExpression,
    SqlDateTruncExpression,
    SqlTimeDeltaExpression,
)
from metricflow.time.time_granularity import TimeGranularity


class BigQuerySqlExpressionRenderer(DefaultSqlExpressionRenderer):
    """Expression renderer for the BigQuery engine."""

    @property
    def double_data_type(self) -> str:
        """Custom double data type for BigQuery engine"""
        return "FLOAT64"

    def visit_cast_to_timestamp_expr(self, node: SqlCastToTimestampExpression) -> SqlExpressionRenderResult:
        """Casts the time value expression to DATETIME, as per standard BigQuery preferences."""
        arg_rendered = self.render_sql_expr(node.arg)
        return SqlExpressionRenderResult(
            sql=f"CAST({arg_rendered.sql} AS DATETIME)",
            execution_parameters=arg_rendered.execution_parameters,
        )

    def visit_date_trunc_expr(self, node: SqlDateTruncExpression) -> SqlExpressionRenderResult:
        """Render DATE_TRUNC for BigQuery, which takes the opposite argument order from Snowflake and Redshift"""
        arg_rendered = self.render_sql_expr(node.arg)

        prefix = ""
        if node.time_granularity in (TimeGranularity.WEEK, TimeGranularity.YEAR):
            prefix = "iso"

        return SqlExpressionRenderResult(
            sql=f"DATE_TRUNC({arg_rendered.sql}, {prefix}{node.time_granularity.value})",
            execution_parameters=arg_rendered.execution_parameters,
        )

    def visit_time_delta_expr(self, node: SqlTimeDeltaExpression) -> SqlExpressionRenderResult:  # noqa: D
        column = node.arg.accept(self)
        if node.grain_to_date:
            granularity = node.granularity
            if granularity == TimeGranularity.WEEK or granularity == TimeGranularity.YEAR:
                granularity.value = "ISO" + granularity.value.upper()
            return SqlExpressionRenderResult(
                sql=f"DATE_TRUNC({column.sql}, {granularity.value})",
                execution_parameters=column.execution_parameters,
            )

        return SqlExpressionRenderResult(
            sql=f"DATE_SUB(CAST({column.sql} AS DATETIME), INTERVAL {node.count} {node.granularity.value})",
            execution_parameters=column.execution_parameters,
        )


class BigQuerySqlQueryPlanRenderer(DefaultSqlQueryPlanRenderer):
    """Plan renderer for the BigQuery engine."""

    EXPR_RENDERER = BigQuerySqlExpressionRenderer()

    @property
    def expr_renderer(self) -> SqlExpressionRenderer:  # noqa :D
        return self.EXPR_RENDERER



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['metricflow',
 'metricflow.api',
 'metricflow.cli',
 'metricflow.configuration',
 'metricflow.constraints',
 'metricflow.dag',
 'metricflow.dataflow',
 'metricflow.dataflow.builder',
 'metricflow.dataset',
 'metricflow.engine',
 'metricflow.errors',
 'metricflow.execution',
 'metricflow.inference',
 'metricflow.inference.context',
 'metricflow.inference.renderer',
 'metricflow.inference.rule',
 'metricflow.inference.solver',
 'metricflow.model',
 'metricflow.model.objects',
 'metricflow.model.objects.constraints',
 'metricflow.model.objects.elements',
 'metricflow.model.parsing',
 'metricflow.model.semantics',
 'metricflow.model.transformations',
 'metricflow.model.validations',
 'metricflow.naming',
 'metricflow.plan_conversion',
 'metricflow.protocols',
 'metricflow.query',
 'metricflow.sql',
 'metricflow.sql.optimizer',
 'metricflow.sql.render',
 'metricflow.sql_clients',
 'metricflow.telemetry',
 'metricflow.telemetry.handlers',
 'metricflow.test',
 'metricflow.test.api',
 'metricflow.test.cli',
 'metricflow.test.constraints',
 'metricflow.test.dataflow',
 'metricflow.test.dataflow.builder',
 'metricflow.test.dataset',
 'metricflow.test.examples',
 'metricflow.test.execution',
 'metricflow.test.fixtures',
 'metricflow.test.inference',
 'metricflow.test.inference.context',
 'metricflow.test.inference.renderer',
 'metricflow.test.inference.rule',
 'metricflow.test.inference.solver',
 'metricflow.test.integration',
 'metricflow.test.model',
 'metricflow.test.model.parsing',
 'metricflow.test.model.semantics',
 'metricflow.test.model.transformations',
 'metricflow.test.model.validations',
 'metricflow.test.plan_conversion',
 'metricflow.test.plan_conversion.dataflow_to_sql',
 'metricflow.test.plan_conversion.instance_converters',
 'metricflow.test.query',
 'metricflow.test.sql',
 'metricflow.test.sql.optimizer',
 'metricflow.test.sql_clients',
 'metricflow.test.telemetry',
 'metricflow.test.time',
 'metricflow.time']

package_data = \
{'': ['*'],
 'metricflow.cli': ['sample_models/*'],
 'metricflow.model.parsing': ['schemas/*'],
 'metricflow.test': ['snapshots/test_column_pruner.py/SqlQueryPlan/*',
                     'snapshots/test_convert_data_source.py/SqlQueryPlan/BigQuerySqlClient/*',
                     'snapshots/test_convert_data_source.py/SqlQueryPlan/DatabricksSqlClient/*',
                     'snapshots/test_convert_data_source.py/SqlQueryPlan/DuckDbSqlClient/*',
                     'snapshots/test_convert_data_source.py/SqlQueryPlan/PostgresSqlClient/*',
                     'snapshots/test_convert_data_source.py/SqlQueryPlan/RedshiftSqlClient/*',
                     'snapshots/test_convert_data_source.py/SqlQueryPlan/SnowflakeSqlClient/*',
                     'snapshots/test_data_warehouse_tasks.py/data_warehouse_validation_model/BigQuerySqlClient/*',
                     'snapshots/test_data_warehouse_tasks.py/data_warehouse_validation_model/DatabricksSqlClient/*',
                     'snapshots/test_data_warehouse_tasks.py/data_warehouse_validation_model/DuckDbSqlClient/*',
                     'snapshots/test_data_warehouse_tasks.py/data_warehouse_validation_model/PostgresSqlClient/*',
                     'snapshots/test_data_warehouse_tasks.py/data_warehouse_validation_model/RedshiftSqlClient/*',
                     'snapshots/test_data_warehouse_tasks.py/data_warehouse_validation_model/SnowflakeSqlClient/*',
                     'snapshots/test_data_warehouse_tasks.py/data_warehouse_validation_model/SqliteSqlClient/*',
                     'snapshots/test_dataflow_plan_builder.py/DataflowPlan/*',
                     'snapshots/test_dataflow_to_execution.py/ExecutionPlan/BigQuerySqlClient/*',
                     'snapshots/test_dataflow_to_execution.py/ExecutionPlan/DatabricksSqlClient/*',
                     'snapshots/test_dataflow_to_execution.py/ExecutionPlan/DuckDbSqlClient/*',
                     'snapshots/test_dataflow_to_execution.py/ExecutionPlan/PostgresSqlClient/*',
                     'snapshots/test_dataflow_to_execution.py/ExecutionPlan/RedshiftSqlClient/*',
                     'snapshots/test_dataflow_to_execution.py/ExecutionPlan/SnowflakeSqlClient/*',
                     'snapshots/test_dataflow_to_execution.py/ExecutionPlan/SqliteSqlClient/*',
                     'snapshots/test_dataflow_to_sql_plan.py/DataflowPlan/*',
                     'snapshots/test_dataflow_to_sql_plan.py/SqlQueryPlan/*',
                     'snapshots/test_dataflow_to_sql_plan.py/SqlQueryPlan/BigQuerySqlClient/*',
                     'snapshots/test_dataflow_to_sql_plan.py/SqlQueryPlan/DatabricksSqlClient/*',
                     'snapshots/test_dataflow_to_sql_plan.py/SqlQueryPlan/DuckDbSqlClient/*',
                     'snapshots/test_dataflow_to_sql_plan.py/SqlQueryPlan/PostgresSqlClient/*',
                     'snapshots/test_dataflow_to_sql_plan.py/SqlQueryPlan/RedshiftSqlClient/*',
                     'snapshots/test_dataflow_to_sql_plan.py/SqlQueryPlan/SnowflakeSqlClient/*',
                     'snapshots/test_dataflow_to_sql_plan.py/SqlQueryPlan/SqliteSqlClient/*',
                     'snapshots/test_engine_specific_rendering.py/SqlQueryPlan/BigQuerySqlClient/*',
                     'snapshots/test_engine_specific_rendering.py/SqlQueryPlan/DatabricksSqlClient/*',
                     'snapshots/test_engine_specific_rendering.py/SqlQueryPlan/DuckDbSqlClient/*',
                     'snapshots/test_engine_specific_rendering.py/SqlQueryPlan/PostgresSqlClient/*',
                     'snapshots/test_engine_specific_rendering.py/SqlQueryPlan/RedshiftSqlClient/*',
                     'snapshots/test_engine_specific_rendering.py/SqlQueryPlan/SnowflakeSqlClient/*',
                     'snapshots/test_metric_time_dimension_to_sql.py/SqlQueryPlan/*',
                     'snapshots/test_metric_time_dimension_to_sql.py/SqlQueryPlan/BigQuerySqlClient/*',
                     'snapshots/test_metric_time_dimension_to_sql.py/SqlQueryPlan/DatabricksSqlClient/*',
                     'snapshots/test_metric_time_dimension_to_sql.py/SqlQueryPlan/DuckDbSqlClient/*',
                     'snapshots/test_metric_time_dimension_to_sql.py/SqlQueryPlan/PostgresSqlClient/*',
                     'snapshots/test_metric_time_dimension_to_sql.py/SqlQueryPlan/RedshiftSqlClient/*',
                     'snapshots/test_metric_time_dimension_to_sql.py/SqlQueryPlan/SnowflakeSqlClient/*',
                     'snapshots/test_metric_time_dimension_to_sql.py/SqlQueryPlan/SqliteSqlClient/*',
                     'snapshots/test_plot_time_dimension_to_sql.py/SqlQueryPlan/*',
                     'snapshots/test_plot_time_dimension_to_sql.py/SqlQueryPlan/SqliteSqlClient/*',
                     'snapshots/test_rendered_query.py/MetricFlowExplainResult/BigQuerySqlClient/*',
                     'snapshots/test_rendered_query.py/MetricFlowExplainResult/DatabricksSqlClient/*',
                     'snapshots/test_rendered_query.py/MetricFlowExplainResult/DuckDbSqlClient/*',
                     'snapshots/test_rendered_query.py/MetricFlowExplainResult/PostgresSqlClient/*',
                     'snapshots/test_rendered_query.py/MetricFlowExplainResult/RedshiftSqlClient/*',
                     'snapshots/test_rendered_query.py/MetricFlowExplainResult/SnowflakeSqlClient/*',
                     'snapshots/test_rendered_query.py/MetricFlowExplainResult/SqliteSqlClient/*',
                     'snapshots/test_rewriting_sub_query_reducer.py/SqlQueryPlan/*',
                     'snapshots/test_sql_plan_render.py/SqlQueryPlan/BigQuerySqlClient/*',
                     'snapshots/test_sql_plan_render.py/SqlQueryPlan/DatabricksSqlClient/*',
                     'snapshots/test_sql_plan_render.py/SqlQueryPlan/DuckDbSqlClient/*',
                     'snapshots/test_sql_plan_render.py/SqlQueryPlan/PostgresSqlClient/*',
                     'snapshots/test_sql_plan_render.py/SqlQueryPlan/RedshiftSqlClient/*',
                     'snapshots/test_sql_plan_render.py/SqlQueryPlan/SnowflakeSqlClient/*',
                     'snapshots/test_sub_query_reducer.py/SqlQueryPlan/*',
                     'snapshots/test_table_alias_simplifier.py/SqlQueryPlan/*'],
 'metricflow.test.fixtures': ['model_yamls/composite_identifier_model/*',
                              'model_yamls/config_linter_model/*',
                              'model_yamls/data_warehouse_validation_model/data_sources/*',
                              'model_yamls/extended_date_model/data_sources/*',
                              'model_yamls/join_types_model/*',
                              'model_yamls/multi_hop_join_model/partitioned_data_sources/*',
                              'model_yamls/multi_hop_join_model/unpartitioned_data_sources/*',
                              'model_yamls/non_ds_model/*',
                              'model_yamls/simple_model/*',
                              'model_yamls/simple_model/data_sources/*'],
 'metricflow.test.integration': ['test_cases/*']}

install_requires = \
['GitPython>=3.1.27,<4.0.0',
 'Jinja2>=2.11.3',
 'MarkupSafe==2.0.1',
 'PyYAML>=6.0,<7.0',
 'SQLAlchemy>=1.4.27,<2.0.0',
 'click>=7.1.2',
 'croniter>=1.3.4,<2.0.0',
 'databricks-sql-connector==2.0.3',
 'duckdb-engine>=0.1.8,<0.2.0',
 'duckdb==0.3.4',
 'fuzzywuzzy>=0.18.0,<0.19.0',
 'graphviz==0.18.2',
 'halo>=0.0.31,<0.0.32',
 'jsonschema==3.2.0',
 'more-itertools==8.10.0',
 'moz-sql-parser>=4.40.21126,<5.0.0',
 'numpy>=1.22.2',
 'pandas>=1.3.0,<2.0.0',
 'pycron>=3.0.0,<4.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'python-Levenshtein>=0.12.2,<0.13.0',
 'python-dateutil==2.8.2',
 'requests>=2.27.1,<3.0.0',
 'ruamel.yaml>=0.17.21,<0.18.0',
 'rudder-sdk-python>=1.0.3,<2.0.0',
 'snowflake-sqlalchemy==1.2.3',
 'sqlalchemy-bigquery>=1.4.3,<2.0.0',
 'sqlalchemy-redshift==0.8.1',
 'sqlalchemy2-stubs>=0.0.2-alpha.21,<0.0.3',
 'tabulate==0.8.9',
 'typing_extensions>=4.0.0',
 'update-checker>=0.18.0,<0.19.0',
 'yamllint>=1.26.3,<2.0.0']

extras_require = \
{'dbt-bigquery': ['dbt-bigquery>=1.3.0,<2.0.0'],
 'dbt-postgres': ['dbt-postgres>=1.3.0,<2.0.0'],
 'dbt-redshift': ['dbt-redshift>=1.3.0,<2.0.0'],
 'dbt-snowflake': ['dbt-snowflake>=1.3.0,<2.0.0']}

entry_points = \
{'console_scripts': ['mf-lite = metricflow.cli.main:cli']}

setup_kwargs = {
    'name': 'metricflow-lite',
    'version': '0.130.1',
    'description': 'Lite version of MetricFlow. This version is missing dependencies for using connectors with data warehouses',
    'long_description': '**Note**\nThis is not the full [MetricFlow](https://github.com/transform-data/metricflow/) package, this is MetricFlow-Lite. MetricFlow-Lite is a forked version of MetricFlow with some deps removed.\n\n**Warning**\nMetricFlow-Lite is not supported. I.e., issues/bugs/change requests should happen on the full [MetricFlow](https://github.com/transform-data/metricflow/) pacakge. \n\n# Welcome to MetricFlow-Lite\n\nMetricFlow-Lite is a forked version of MetricFlow with data warehouse connector dependencies removed. This necessary for some of our (Transform) internal projects. \n\n## License\n\nMetricFlow-Lite is open source software. The project relies on several licenses including AGPL-3.0-or-later and Apache (specified at folder level).\n\nMetricFlow-Lite is built by [Transform](https://transform.co/)\n',
    'author': 'Transform',
    'author_email': 'hello@transformdata.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://transform.co/metricflow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)
